/*****************************************************************************
 * game.cpp - General game loop class
 *
 *****************************************************************************/
#include "SDL.h"

#include "game.h"
#include "freecnc.h" // for the game definitions
#include "vqa.h"
#include "memtrack.h"

int Game::mapscaleq = -1;
/** Constructor, loads the map, sidebar and such. plays briefingand actionmovie
 * @param the graphicsengine.
 * @param the soundengine.
 * @param the name of the mission to play.
 */
Game::Game( GraphicsEngine *gfxengine,
            SoundEngine *sndengine, char *missionname, Uint8 gamenum ) {
    VQAMovie *mov;
    char* message,*tmp;
    INIFile* fileini;

    /* if the game is the demo, revert it back to the
     * original TD as everything is the same now
     */
    gamenum = gamenum == GAME_TD_DEMO ? GAME_TD : gamenum;
    /* set the pointer to the gfx engine */
    this->gfxengine = gfxengine;
    this->sndengine = sndengine;
    try {
        fileini = new INIFile("files.ini");
    } catch(INInotFound *e) {
        delete e;
        return;
    }
    tmp = new char[8];
    sprintf(tmp,"game%2i",gamenum);
    message = fileini->readString("general",tmp,"TD");
    delete[] tmp;
    if (!sndengine->createPlayList(message)) {
        exit(1);
    }
    delete[] message;
    delete fileini;

    /* reset the tickcounter, should be a function in the class conatining the
     * counter */
    fprintf(stderr,"\tCreating the ActionEventQueue\n");
    evQueue = new ActionEventQueue();
    /* load the map */
    fprintf(stderr,"\tLoading the map (this might take a while)\n");
    map = new CnCMap(evQueue, missionname, gamenum);
    map->getUASP()->setsfxeng(sndengine);
    /* play briefing */
    try {
        mov = new VQAMovie(gfxengine, sndengine, map->getBrief());
        mov->play();
        delete mov;
    } catch (VQAError *vqaerr) {
        delete vqaerr;
    }
    try {
        mov = new VQAMovie(gfxengine, sndengine, map->getAction());
        mov->play();
        delete mov;
    } catch (VQAError *vqaerr) {
        delete vqaerr;

    }
    /* init sidebar */
    fprintf(stderr,"\tInitialising the sidebar\n");
    sidebar = new Sidebar(evQueue, map->getPlayerPool()->getHPlayer(),
                          gfxengine->getHeight(), map->getTheater());
    /* init cursor */
    fprintf(stderr,"\tInitialising the cursor\n");
    cursor = new Cursor();
    /* init the input functions */
    fprintf(stderr,"\tInitialising the input engine\n");
    input = new Input(sidebar, map, gfxengine->getWidth(),
                      gfxengine->getHeight(), gfxengine->getMapArea(), cursor, sndengine);
    //sprintf(message, "Loaded map \"%s\"", missionname);
    //gfxengine->postMessage(message);
}

/** Destructor, frees up some memory */
Game::~Game() {
    delete evQueue;
    delete input;
    delete sidebar;
    delete map;
    delete cursor;
}

/** Play the mission. */
void Game::play() {
    VQAMovie *mov;
    /* start the song */
    SDL_LockAudio();
    sndengine->playSong(map->getTheme());
    SDL_UnlockAudio();
    /* send a pointer to the map, sidebar and cursor to the gfxengine */
    gfxengine->setupCurrentGame(map, sidebar, cursor);
    /* main gameloop */
    //   gfxengine->postMessage("Starting main gameloop..");
    while( !input->shouldQuit() ) {
        /* draw the scene */
        gfxengine->renderScene();
        /* run all sceduled events */
        evQueue->runEvents();
        /* handle the input */
        input->handle();
    }
    /* stop the song */
    sndengine->stopSong();

    if( map->getPlayerPool()->hasWon() ) {
        try {
            mov = new VQAMovie(gfxengine, sndengine, map->getWinMov());
            mov->play();
            delete mov;
        } catch (VQAError *vqaerr) {
            delete vqaerr;
        }
    } else if( map->getPlayerPool()->hasLost() ) {
        try {
            mov = new VQAMovie(gfxengine, sndengine, map->getLoseMov());
            mov->play();
            delete mov;
        } catch (VQAError *vqaerr) {
            delete vqaerr;
        }
    }
}
/*
class FooEvent : public ActionEvent{
 public:
   FooEvent( Uint32 p, Game *g ) : ActionEvent(p){
      game = g;
   }
   virtual void run() const {
      game->foo();
   }
 private:
   Game *game;
};
 
void Game::foo()
{
//   gp = this;
   evQueue->scheduleEvent(new FooEvent(300, this));
   printf("hej\n");
}
 
*/
