/*****************************************************************************
 * path.cpp - A path finder and representation
 *
 ****************************************************************************/
#include "SDL.h"
#include <stdlib.h>


#include "path.h"



/* temporary code since the fibheap isn't working yet */
#include <queue>
#include <vector>

#include "memtrack.h"


using namespace std;


//#include "fibheap.h"

//Uint16 Path::mapwidth;
//Uint16 Path::mapheight;
//Uint16 Path::mapsize;

#define Max( x, y ) (x>y?x:y)
#define Min( x, y ) (x>y?y:x)

typedef struct {
    Uint16 cr;
    Uint32 g;
    Uint32 h;
}
TileRef;

class FibHeapEntry {
public:
    friend class KeyComp;
    FibHeapEntry(TileRef *val, Uint32 k) {
        value = val;
        key = k;
    }
    TileRef *getValue() {
        return value;
    }
    void setKey(Uint32 k) {
        key = k;
    }
private:
    TileRef *value;
    Uint32 key;
};
// Friend class which compares ActionEvents priority
class KeyComp {
public:
    bool operator()(FibHeapEntry *x, FibHeapEntry *y) {
        return x->key > y->key;
    }
};


//Path::Path( Uint16 crBeg, Uint16 crEnd, Uint16 *edgecostmatrix ) : stack<Uint8>()
Path::Path( Uint16 crBeg, Uint16 crEnd, CnCMap *ccmap, Uint8 max_dist ) : stack<Uint8>() {
    //   Uint16 crBeg = startposy*mapwidth + startposx;
    Uint16 startposx = crBeg%ccmap->getWidth();
    Uint16 startposy = crBeg/ccmap->getWidth();
    //   Uint16 crEnd = stopposy*mapwidth + stopposx;
    Uint16 stopposx = crEnd%ccmap->getWidth();
    Uint16 stopposy = crEnd/ccmap->getWidth();
    FibHeapEntry** Nodes;// = new (FibHeapEntry*)[mapsize];
    Uint8 *ed;// = new Uint8[mapsize];
    FibHeapEntry *pU, *pV;
    FibHeapEntry *pReuse = NULL;
    //   FibHeap PQ;
    std::priority_queue<FibHeapEntry*, std::vector<FibHeapEntry *>, KeyComp> PQ;
    Uint16 crU, crV;
    Uint32 dwCost;
    int i;
    TileRef *cellTempU, *cellTempV;
    Uint16 diffx, diffy;
    Uint8 edp;
    Uint16 cmincr;
    Uint32 cminh;

    Uint16 mapsize = ccmap->getWidth()*ccmap->getHeight();

    FibHeapEntry *retired;

    //    if( ccmap->getCost(crEnd) >= 0xfff0 ) {
    //        return;
    //    }


    Nodes = new FibHeapEntry*[mapsize];

    ed = new Uint8[mapsize];


    for( i = 0; i < mapsize; i++ )
        Nodes[i] = NULL;

    cellTempU = new TileRef;
    cellTempU->cr = crBeg;
    cellTempU->g = 0;
    diffx = abs( startposx - stopposx );
    diffy = abs( startposy - stopposy );
    cellTempU->h = Min(diffx, diffy)*14 + abs( diffx - diffy )*10;

    Nodes[crBeg] = pU = new FibHeapEntry(cellTempU, cellTempU->h);
    cmincr = cellTempU->cr;
    cminh = cellTempU->h;

    PQ.push(pU);

    retired = new FibHeapEntry(NULL, 0);

    /* If there are more tiles to visit, visit them */
    //   while( (pU = PQ.pop()) != NULL ){
    while( !PQ.empty() ) {
        pU = PQ.top();
        PQ.pop();
        cellTempU = (TileRef *)(pU->getValue());
        crU = cellTempU->cr;

        /* Break if we are at the destination point */
        //      if( crU == crEnd ){
        //	 dwCost = cellTempU->g;
        //	 delete cellTempU;
        //	 delete pU;
        //	 Nodes[crU] = NULL;
        //	 break;
        //      }
        if( cellTempU->h < cminh ) {
            cminh = cellTempU->h;
            cmincr = cellTempU->cr;
        }

        if( cellTempU->h <= ((Uint32)max_dist)*10 ) /* desired min dist to target, 0 to go all the way */
            break;

        if( cellTempU->g >= 100 ) {
            break;
        }



        /* Walk in all directions */
        for( edp = 0; edp < 8; edp++ ) {
            switch( edp ) {
            case 0:
                if( (crU / ccmap->getWidth()) == 0 )
                    dwCost = 0xffff;
                else {
                    crV = crU - ccmap->getWidth();
                    if( ccmap->getCost(crV) == 0xffff )
                        dwCost = 0xffff;
                    else
                        dwCost = cellTempU->g + ccmap->getCost(crV)*10;
                }
                break;
            case 1:
                if( (crU / ccmap->getWidth()) == 0 || (crU % ccmap->getWidth()) == (ccmap->getWidth()-1) )
                    dwCost = 0xffff;
                else {
                    crV = crU - ccmap->getWidth() + 1;
                    if( ccmap->getCost(crV) == 0xffff )
                        dwCost = 0xffff;
                    else
                        dwCost = cellTempU->g + ccmap->getCost(crV)*14;
                }
                break;
            case 2:
                if( (crU % ccmap->getWidth()) == (ccmap->getWidth()-1) )
                    dwCost = 0xffff;
                else {
                    crV = crU + 1;
                    if( ccmap->getCost(crV) == 0xffff )
                        dwCost = 0xffff;
                    else
                        dwCost = cellTempU->g + ccmap->getCost(crV)*10;
                }
                break;
            case 3:
                if( (crU / ccmap->getWidth()) == (ccmap->getHeight()-1) ||
                        (crU % ccmap->getWidth()) == (ccmap->getWidth()-1) )
                    dwCost = 0xffff;
                else {
                    crV = crU + ccmap->getWidth() + 1;
                    if( ccmap->getCost(crV) == 0xffff )
                        dwCost = 0xffff;
                    else
                        dwCost = cellTempU->g + ccmap->getCost(crV)*14;
                }
                break;
            case 4:
                if( (crU / ccmap->getWidth()) == (ccmap->getHeight()-1) )
                    dwCost = 0xffff;
                else {
                    crV = crU + ccmap->getWidth();
                    if( ccmap->getCost(crV) == 0xffff )
                        dwCost = 0xffff;
                    else
                        dwCost = cellTempU->g + ccmap->getCost(crV)*10;
                }
                break;
            case 5:
                if( (crU / ccmap->getWidth()) == (ccmap->getHeight()-1) || (crU % ccmap->getWidth()) == 0 )
                    dwCost = 0xffff;
                else {
                    crV = crU + ccmap->getWidth() - 1;
                    if( ccmap->getCost(crV) == 0xffff )
                        dwCost = 0xffff;
                    else
                        dwCost = cellTempU->g + ccmap->getCost(crV)*14;
                }
                break;
            case 6:
                if( (crU % ccmap->getWidth()) == 0 )
                    dwCost = 0xffff;
                else {
                    crV = crU - 1;
                    if( ccmap->getCost(crV) == 0xffff )
                        dwCost = 0xffff;
                    else
                        dwCost = cellTempU->g + ccmap->getCost(crV)*10;
                }
                break;
            case 7:
                if( (crU / ccmap->getWidth()) == 0 || (crU % ccmap->getWidth()) == 0 )
                    dwCost = 0xffff;
                else {
                    crV = crU - ccmap->getWidth() - 1;
                    if( ccmap->getCost(crV) == 0xffff )
                        dwCost = 0xffff;
                    else
                        dwCost = cellTempU->g + ccmap->getCost(crV)*14;
                }
                break;
            default:
                break;
            }

            if( dwCost >= 0xffff )
                continue;

            pV = Nodes[crV];

            /* Retired */
            if( pV == retired )
                continue;
            /* Not visited */
            else if( pV == NULL ) {
                if( pReuse != NULL ) {
                    pV = pReuse;
                    pReuse = NULL;
                    cellTempV = (TileRef *)(pV->getValue());
                } else {
                    cellTempV = new TileRef;
                    pV = new FibHeapEntry(cellTempV, 0);
                }

                Nodes[crV] = pV;
                ed[crV] = edp;

                cellTempV->cr = crV;
                cellTempV->g = dwCost;
                diffx = abs( crV%ccmap->getWidth() - stopposx );
                diffy = abs( crV/ccmap->getWidth() - stopposy );
                cellTempV->h = Min(diffx, diffy)*14 + abs( diffx - diffy )*10;

                pV->setKey(cellTempV->g + cellTempV->h);

                PQ.push(pV);
            } else if( dwCost <= ((TileRef *)(pV->getValue()))->g ) {
                if( dwCost == ((TileRef *)(pV->getValue()))->g ) {
                    //	       if( rand() <= RAND_MAX/2 )
                    //		 ed[crV] = edp;
                    continue;
                }
                ((TileRef *)(pV->getValue()))->g = dwCost;
                ed[crV] = edp;
                /* Decrease the key */
                //	    PQ.decreaseKey( pV, dwCost + ((TileRef *)(pV->getValue()))->h );
                pV->setKey(dwCost + ((TileRef *)(pV->getValue()))->h);
            }

        }

        Nodes[crU] = retired;
        if( pReuse != NULL ) {
            delete cellTempU;
            delete pU;
        } else
            pReuse = pU;

    } /* while done */

    if( crU != crEnd ) {
        crEnd = cmincr;
        crU = crEnd;
    }

    if( pReuse != NULL ) {
        delete (TileRef *)pReuse->getValue();
        delete pReuse;
    }

    for( i = 0; i < mapsize; i++ )
        if( Nodes[i] != NULL && Nodes[i] != retired ) {
            delete (TileRef *)Nodes[i]->getValue();
            delete Nodes[i];
        }
    delete[] Nodes;
    delete retired;


    if( crEnd != crBeg ) {
        //      path = createStack();
        //
        crV = crEnd;

        while( crV != crBeg ) {
            //	 path = addValueStack(path, ed[crV]);
            push(ed[crV]);
            switch( ed[crV] ) {
            case 0:
                crV = crV + ccmap->getWidth();
                break;
            case 1:
                crV = crV + ccmap->getWidth() - 1;
                break;
            case 2:
                crV = crV - 1;
                break;
            case 3:
                crV = crV - ccmap->getWidth() - 1;
                break;
            case 4:
                crV = crV - ccmap->getWidth();
                break;
            case 5:
                crV = crV - ccmap->getWidth() + 1;
                break;
            case 6:
                crV = crV + 1;
                break;
            case 7:
                crV = crV + ccmap->getWidth() + 1;
                break;
            default:
                delete[] ed;
                //	    delete this;
                return;
            }
        }
        delete[] ed;
        return;
    } else {
        delete[] ed;
        //     delete this;
        return;
    }
}

Path::~Path() {
    while( !empty() )
        pop();
}
