#include <string.h>
#include "playerpool.h"
#include "inifile.h"

#include "message.h"

#include "memtrack.h"
#include "ai.h"

#ifndef __GNUC__
int strcasecmp(char*, char*);
#endif

using namespace std;

Player::Player(char *pname, INIFile *mapini, PlayerPool *ppool) {
    playername = new char[strlen(pname)+1];
    strcpy(playername, pname);
    if( !strcasecmp(playername, "goodguy") ) {
        playerside = PS_GOOD;
        palnum = 0;
    } else if( !strcasecmp(playername, "badguy") ) {
        playerside = PS_BAD;
        palnum = 1;
    } else if( !strcasecmp(playername, "neutral") ) {
        playerside = PS_NEUTRAL;
        palnum = 0;
    } else if( !strcasecmp(playername, "special") ) {
        playerside = PS_SPECIAL;
        palnum = 0;
    } else {
        fprintf(stderr, "Player Side \"%s\" not recognised, using gdi instead\n",pname);
        playerside = PS_GOOD;
        palnum = 0;
    }
    money = 0;
    pwr = 0;
    pwrused = 0;
    numallies = 0;
    this->ppool = ppool;
    myAI = NULL;
}

Player::~Player() {
    delete[] playername;
}

bool Player::decreaseMoney(Uint32 change) {
    if( money < change ) {
        return false;
    }
    money -= change;
    return true;
}

void Player::builtUnit(Unit* un) {
    unitpool.push_back(un);
}

void Player::lostUnit(Unit* un, bool wasDeployed) {
    char left[1024];
    Uint32 i;
    if (!wasDeployed) {
        sprintf(left, "%s has %d structs and %d units", playername, structurepool.size(), unitpool.size()-1);
        MessagePool::currentInstance->postMessage(left);
        if (myAI != NULL) {
            myAI->lostUnit(un);
        }
    }
    if( unitpool.size() <= 1 && structurepool.size() == 0 && !wasDeployed) {
        sprintf(left, "Player \"%s\" defeated", playername);
        MessagePool::currentInstance->postMessage(left);
        ppool->playerDefeated(this);
    } else {
        for (i=0;i<unitpool.size();++i) {
            if (unitpool[i] == un) {
                break;
            }
        }
        for (i=i+1;i<unitpool.size();++i) {
            unitpool[i-1] = unitpool[i];
        }
        unitpool.resize(unitpool.size()-1);
    }
}

void Player::builtStruct(Structure* str) {
    structurepool.push_back(str);
    powerinfo_t newpower = ((StructureType*)str->getType())->getPowerInfo();
    pwr += newpower.power;
    pwrused += newpower.drain;
}

void Player::lostStruct(Structure* str) {
    Uint32 i;
    char left[1024];
    powerinfo_t newpower = ((StructureType*)str->getType())->getPowerInfo();
    pwr -= newpower.power;
    pwrused -= newpower.drain;
    sprintf(left, "%s has %d structs and %d units", playername, structurepool.size()-1, unitpool.size());
    MessagePool::currentInstance->postMessage(left);
    if (myAI != NULL) {
        myAI->lostStructure(str);
    }
    if( unitpool.size() == 0 && structurepool.size() <= 1 ) {
        sprintf(left, "Player \"%s\" defeated", playername);
        MessagePool::currentInstance->postMessage(left);
        ppool->playerDefeated(this);
    } else {
        for (i=0;i<structurepool.size();++i) {
            if (structurepool[i] == str) {
                break;
            }
        }
        for (i=i+1;i<structurepool.size();++i) {
            structurepool[i-1] = structurepool[i];
        }
        structurepool.resize(structurepool.size()-1);
    }
}

bool Player::isAllied() {
    return false;
}

void Player::createAI(CnCMap *theMap) {
    myAI = new AI(theMap, this);
}

void Player::updateOwner(Uint8 newnum) {
    Uint32 i;
    for (i=0;i<unitpool.size();++i)
        unitpool[i]->setOwner(newnum);
    for (i=0;i<structurepool.size();++i)
        structurepool[i]->setOwner(newnum);
}

PlayerPool::PlayerPool() {
    lost = false;
    won = false;
}

PlayerPool::~PlayerPool() {
    Uint8 i;

    for( i = 0; i < playerpool.size(); i++ ) {
        delete playerpool[i];
    }
}

void PlayerPool::setHPlayer(char *pname) {
    Uint8 i;
    for( i = 0; i < playerpool.size(); i++ ) {
        if( !strcasecmp(playerpool[i]->getName(), pname) ) {
            humanPlayer = i;
            return;
        }
    }
    fprintf(stderr, "Tried to set human player to non-existing player \"%s\"\n", pname);
    playerpool.push_back(new Player(pname, NULL, this));
    humanPlayer = playerpool.size()-1;
    //    exit(1);
}

Uint8 PlayerPool::getPlayerNum(char *pname, INIFile *mapini) {
    Uint8 i;
    for( i = 0; i < playerpool.size(); i++ ) {
        if( !strcasecmp(playerpool[i]->getName(), pname) ) {
            return i;
        }
    }
    playerpool.push_back(new Player(pname, mapini, this));
    return playerpool.size()-1;
}

Uint8 PlayerPool::getPlayerNum(Player *thePlayer) {
    for(Uint8 i = 0; i < playerpool.size(); i++ ) {
        if (playerpool[i] == thePlayer) {
            return i;
        }
    }
    return 255;
}

vector<Uint8> PlayerPool::getOwnersForSide(PSIDE thePlayerSide) {
    vector<Uint8> aOwners(0);
    for(Uint8 i = 0; i < playerpool.size(); i++ ) {
        if (playerpool[i]->getSide() == thePlayerSide) {
            aOwners.push_back(i);
        }
    }
    return aOwners;
}

void PlayerPool::playerDefeated(Player *pl) {
    Uint8 i;

    for( i = 0; i < playerpool.size(); i++ ) {
        if( playerpool[i] == pl ) {
            break;
        }
    }
    if( i == humanPlayer ) {
        lost = true;
    }
    if( i < humanPlayer ) {
        humanPlayer--;
    }
    for(i = i+1;i < playerpool.size(); i++ ) {
        playerpool[i-1] = playerpool[i];
    }
    playerpool.resize(playerpool.size()-1);
    for (i=0;i<playerpool.size();++i)
        playerpool[i]->updateOwner(i);

    if( !lost ) {
        won = true;
        for( i = 0; i < playerpool.size() ; i++ ) {
            if( i != humanPlayer && playerpool[i]->getSide() != PS_NEUTRAL ) {
                won = false;
            }
        }
    }

}
