#include <stdlib.h>
#include <string.h>
#include "structure.h"
#include "inifile.h"

#include "game.h"

#include "memtrack.h"

#include "weaponspool.h"

using namespace std;

#ifndef __GNUC__
int strcasecmp(char*, char*);
#endif

UnitAndStructurePool *Structure::uspool;

StructureType::StructureType(vector<SHPImage *> *imgPool, char *tname, INIFile *structini,
                             char *thext, WeaponsPool* wpool) {
    SHPImage *shpimage, *makeimage;
    int i;
    char shpname[13], imagename[8];
    char* tmp;
    INIFile *arts;
    char blocktest[128];
    Uint32 size;
    char* miscnames;

    is_wall = false;

    if( !strcasecmp(tname, "BRIK") || !strcasecmp(tname, "SBAG") ||
            !strcasecmp(tname, "BARB") || !strcasecmp(tname, "WOOD") ||
            !strcasecmp(tname, "CYCL") )
        is_wall = true;

    /* the size of the structure in tiles */
    arts = new INIFile("art.ini");
    xsize = arts->readInt(tname, "xsize",1);
    ysize = arts->readInt(tname, "ysize",1);

    blckoff = 0;
    blocked = new Uint8[xsize*ysize];
    for( i = 0; i < xsize*ysize; i++ ) {
        sprintf(blocktest, "notblocked%d", i);
        size = arts->readInt(tname, blocktest);
        if(size == INIERROR) {
            blocked[i] = 1;
            xoffset = -(i%xsize)*24;
            yoffset = -(i/xsize)*24;
            if (blocked[blckoff] == 0) {
                blckoff = i;
            }
        } else {
            blocked[i] = 0;
        }
    }
    delete arts;

    memset(shpname,0,13);
    numshps = structini->readInt(tname, "layers",1);

    shpnums = new Uint16[(is_wall?numshps:numshps+1)];
    shptnum = new Uint16[numshps];

    powerinfo.power = structini->readInt(tname,"power",0);
    powerinfo.drain = structini->readInt(tname,"drain",0);
    powerinfo.powered = structini->readInt(tname,"powered",0) != 0;
    maxhealth = structini->readInt(tname,"health",100);

    for( i = 0; i < numshps; i++ ) {
        sprintf(imagename, "image%d", i+1);
        tmp = structini->readString(tname, imagename);
        if( tmp == NULL ) {
            strncpy(shpname, tname, 13);
            strncat(shpname, ".SHP", 13);
        } else {
            strncpy(shpname, tmp, 13);
            delete[] tmp;
        }
        try {
            shpimage = new SHPImage( shpname, Game::mapscaleq);
        } catch( ImageNotFound *e ) {
            delete e;
            strncpy(shpname, tname, 13);
            strncat(shpname, thext, 13);
            try {
                shpimage = new SHPImage( shpname, Game::mapscaleq);
            } catch( ImageNotFound *e2 ) {
                delete e2;
                fprintf(stderr, "Image not found: \"%s\"\n", shpname);
                numshps = 0;
                return;
            }
        }
        shpnums[i] = imgPool->size();
        shptnum[i] = shpimage->getNumImg();
        imgPool->push_back(shpimage);
    }
    if (!is_wall) {
        numwalllevels = 0;
        animinfo.loopend = structini->readInt(tname,"loopend",0);
        animinfo.loopend2 = structini->readInt(tname,"loopend2",0);

        animinfo.animspeed = structini->readInt(tname,"animspeed", 3);
        animinfo.animspeed = abs(animinfo.animspeed);
        animinfo.animspeed = (animinfo.animspeed>1?animinfo.animspeed:2);
        animinfo.animdelay = structini->readInt(tname,"delay",0);

        animinfo.animtype = structini->readInt(tname, "animtype", 0);
        animinfo.sectype  = structini->readInt(tname, "sectype", 0);

        animinfo.dmgoff   = structini->readInt(tname, "dmgoff", ((shptnum[0]-1)>>1));
        if (numshps == 2)
            animinfo.dmgoff2 = structini->readInt(tname, "dmgoff2", (shptnum[1]>>1));
        else
            animinfo.dmgoff2 = 0;

        defaultface       = structini->readInt(tname, "defaultface", 0);

        if (strlen(tname) <= 4) {
            strncpy(shpname, tname, 13);
            strncat(shpname, "make.shp", 13);
        } else
            fprintf(stderr,"Warning! %s is nonstandard!\n",tname);
        try {
            makeimage = new SHPImage(shpname, Game::mapscaleq);
            makeimg = imgPool->size();
            animinfo.makenum = makeimage->getNumImg();
            imgPool->push_back(makeimage);
        } catch( ImageNotFound *e3 ) {
            delete e3;
            // fprintf(stderr, "Image not found: \"%s\"\n", shpname);
            makeimg = 0;
            animinfo.makenum = 0;
            //return;
        }

        miscnames = structini->readString(tname, "primary_weapon");
        if( miscnames == NULL ) {
            primary_weapon = NULL;
        } else {
            primary_weapon = wpool->getWeapon(miscnames);
            delete[] miscnames;
        }
        miscnames = structini->readString(tname, "secondary_weapon");
        if( miscnames == NULL ) {
            secondary_weapon = NULL;
        } else {
            secondary_weapon = wpool->getWeapon(miscnames);
            delete[] miscnames;
        }
        turret = (bool)structini->readInt(tname,"turret",0);
        if (turret) {
            turnspeed = structini->readInt(tname,"rot",3);
        }
        miscnames = structini->readString(tname,"armour","none");
        cost = 0;
        if (miscnames == NULL)
            armour = 0;
        else {
            if (strncasecmp(miscnames,"none",4) == 0)
                armour = 0;
            else if (strncasecmp(miscnames,"wood",4) == 0)
                armour = 1;
            else if (strncasecmp(miscnames,"light",5) == 0)
                armour = 2;
            else if (strncasecmp(miscnames,"heavy",5) == 0)
                armour = 3;
            else if (strncasecmp(miscnames,"concrete",8) == 0)
                armour = 4;
        }
    } else {
        numwalllevels = structini->readInt(tname,"levels",1);
    }
}


StructureType::~StructureType() {
    delete[] shpnums;
    delete[] blocked;
    delete[] shptnum;
}

/* pass owner, cellpos and health to this constructor */
Structure::Structure(StructureType *type, Uint16 cellpos, Uint8 owner) {
    int i;
    this->type = type;
    imagenumbers = new Uint16[type->getNumLayers()];
    for(i=0;i<type->getNumLayers();i++) {
        imagenumbers[i] = 0;
        if( owner != 0xff && !type->isWall() ) {
            imagenumbers[i] |= (uspool->getMap()->getPlayerPool()->getPaletteNum(owner)<<11);
        }
    }
    deleted = false;
    exploding = false;
    this->owner = owner;
    references = 0;
    this->cellpos = cellpos;
    bcellpos = cellpos+(type->getBlckOff() % type->getXsize()) +
               ((type->getBlckOff()/type->getXsize())*uspool->getMap()->getWidth());
    selected = false;
    animating = false;
    usemakeimgs = false;
    buildAnim = NULL;
    attackAnim = NULL;
    health = type->getMaxHealth(); // FIXME: read from mission ini if poss
    damaged = checkdamage();
    if( !type->isWall() ) {
        uspool->getMap()->getPlayerPool()->getPlayer(owner)->builtStruct(this);
    }
}

Structure::~Structure() {
    delete[] imagenumbers;
}

void Structure::remove() {
    if( !type->isWall() ) {
        uspool->getMap()->getPlayerPool()->getPlayer(owner)->lostStruct(this);
    }
    deleted = true;
    if( references == 0 )
        delete this;
}


/** Method to get a list of imagenumbers which the renderer will draw.
 * Owner will be added later to change palette */
Uint8 Structure::getImageNums(Uint32 **inums) {
    Uint16 *shps;
    int i;

    shps = type->getSHPNums();

    if (usemakeimgs && (!type->isWall()) && (type->getMakeImg() != 0)) {
        *inums = new Uint32[1];
        (*inums)[0] = (type->getMakeImg()<<16)|imagenumbers[0];
        return 1;
    } else {
        *inums = new Uint32[type->getNumLayers()];
        for(i = 0; i < type->getNumLayers(); i++ ) {
            (*inums)[i] = (shps[i]<<16)|imagenumbers[i];
        }
        return type->getNumLayers();
    }
}

void Structure::runAnim(Uint32 mode) {
    Uint32 speed;
    if (!animating) {
        animating = true;
        if (mode == 0) { // run build anim at const speed
            usemakeimgs = true;
            buildAnim = new BuildAnimEvent(3,this,false);
        } else {
            speed = type->getAnimInfo().animspeed;
            switch (mode&0xf) {
            case 1:
                if (type->getAnimInfo().animtype == 4) {
                    buildAnim = new ProcAnimEvent(speed,this);
                } else {
                    buildAnim = new LoopAnimEvent(speed,this);
                }
                break;
            case 2:
                buildAnim = new BTurnAnimEvent(speed,this,(mode>>4));
                break;
            case 7:
                buildAnim = new RefineAnimEvent(speed,this,5);
                break;
            default:
                buildAnim = NULL;
                animating = false;
                break;
            }
        }
        if (buildAnim != NULL) {
            uspool->scheduleUSEvent(buildAnim);
        }
    }
}

void Structure::runSecAnim(Uint32 param) {
    BuildingAnimEvent* sec_anim;
    Uint8 secmode = type->getAnimInfo().sectype;
    if (secmode != 0) {
        switch (secmode) {
        case 7:
            sec_anim = new RefineAnimEvent(2,this,param);
            break;
        case 5:
            sec_anim = new DoorAnimEvent(2,this,true);
            break;
        case 8:
            //	       sec_anim = new RepairAnimEvent(3,this);
            break;
        }
        if (animating) {
            buildAnim->setSchedule(sec_anim);
            stopAnim();
        } else {
            buildAnim = sec_anim;
            uspool->scheduleUSEvent(buildAnim);
            animating = true;
        }
    }
}

void Structure::stopAnim() {
    buildAnim->stop();
}

void Structure::unrefer() {
    references--;
    if( deleted && references == 0 ) {
        delete this;
    }
}

void Structure::applyDamage(Sint16 amount, Weapon* weap) {
    if (exploding)
        return;
    Uint8 odam = damaged;
    amount = (Sint16)((double)amount * weap->getVersus(type->getArmour()));
    if ((health-amount) <= 0) {
        exploding = true;
        if (uspool->getMap()->getPlayerPool()->getPlayer(owner)->getAI() != NULL) {
            uspool->getMap()->getPlayerPool()->getPlayer(owner)->getAI()->attackOtherSide();
        }
        BExplodeAnimEvent* boom = new BExplodeAnimEvent(1,this);
        if (animating) {
            buildAnim->setSchedule(boom);
            buildAnim->stop();
        } else {
            buildAnim = boom;
            uspool->scheduleUSEvent(boom);
        }
        return;
    } else if ((health-amount)>type->getMaxHealth()) {
        health = type->getMaxHealth();
    } else {
        health -= amount;
    }
    if (animating) {
        buildAnim->updateDamaged();
    } else {
        if (type->isWall()) {
            if ((damaged = checkdamage())) { // This is correct
                if (odam != damaged) {
                    changeImage(0,16);
                }
            }
        } else {
            if ((damaged = checkdamage())) { // This is correct
                if (odam != damaged) { // only play critical damage sound once
                    uspool->getsfxeng()->queueSound("xplobig4.aud");
                    setimage((imagenumbers[0]&~0x800)+type->getAnimInfo().dmgoff,0);
                    if (type->getNumLayers() == 2)
                        setimage((imagenumbers[1]&~0x800)+type->getAnimInfo().dmgoff2,1);
                }
            } else {
                if (odam) {
                    setimage((imagenumbers[0]&~0x800)-type->getAnimInfo().dmgoff,0);
                    if (type->getNumLayers() == 2)
                        setimage((imagenumbers[1]&~0x800)-type->getAnimInfo().dmgoff,1);
                    return;
                }
            }
        }
    }
}

void Structure::attackUnit(Unit* un) {
    BAttackAnimEvent* nextmove;
    if( attackAnim != NULL ) {
        attackAnim->stop();
        nextmove = new BAttackAnimEvent(0, this, un);
        attackAnim->setSchedule(nextmove);
        attackAnim = nextmove;
    } else {
        attackAnim = new BAttackAnimEvent(0, this, un);
        uspool->scheduleUSEvent(attackAnim);
    }
}

void Structure::attackStructure(Structure* st) {
    BAttackAnimEvent* nextmove;
    if( attackAnim != NULL ) {
        attackAnim->stop();
        nextmove = new BAttackAnimEvent(0, this, st);
        attackAnim->setSchedule(nextmove);
        attackAnim = nextmove;
    } else {
        attackAnim = new BAttackAnimEvent(0, this, st);
        uspool->scheduleUSEvent(attackAnim);
    }
}

void Structure::attackUnitOrStructure(UnitOrStructure* uos) {
    BAttackAnimEvent* nextmove;
    if( attackAnim != NULL ) {
        attackAnim->stop();
        nextmove = new BAttackAnimEvent(0, this, uos);
        attackAnim->setSchedule(nextmove);
        attackAnim = nextmove;
    } else {
        attackAnim = new BAttackAnimEvent(0, this, uos);
        uspool->scheduleUSEvent(attackAnim);
    }
}

Uint8 Structure::checkdamage() {
    ratio = ((double)health)/((double)type->getMaxHealth());
    if (type->isWall()) {
        if ((ratio <= 0.33)&&(type->getNumWallLevels() == 3))
            return 2;
        else if ((ratio <= 0.66)&&(type->getNumWallLevels() > 1))
            return 1;
        else
            return 0;
    } else {
        if (ratio <= 0.5)
            return 1;
        else
            return 0;
    }
}
