#include <stdlib.h>
#include <string.h>
#include "unit.h"
#include "unitandstructurepool.h"

#include "game.h"
#include "weaponspool.h"

#include "string.h"

#include "memtrack.h"

using namespace std;

#ifndef __GNUC__
int strncasecmp(char*, char*, int);
#endif

const Sint8 InfantryGroup::unitoffset[10]= {
            /* Theses values have been heavily tested, do NOT change them unless you're
             *        _really_ sure of what you are doing */
            /* X value */
            -13, -19, -7, -19, -7,
            /* Y value */
            -3, -7, -7, 1, 1
        };


UnitAndStructurePool *Unit::uspool;

UnitType::UnitType(vector<SHPImage *> *imgPool, char *tname, INIFile *unitini,
                   WeaponsPool *wpool, UnitAndStructurePool* uspool) {
    SHPImage *shpimage;
    int i;
    char *shpname;
    Uint32 tmpnumshps;
    Uint32 shpnum;
    Uint32 tmpspeed;
    char* talkmode;
    char *miscnames;

    numlayers = unitini->readInt(tname, "layers", 1);

    shpnums = new Uint32[numlayers];

    shpname = new char[strlen(tname)+6];
    strcpy(shpname, tname);
    strcat(shpname, ".SHP");
    try {
        shpimage = new SHPImage(shpname, Game::mapscaleq);
    } catch( ImageNotFound *e ) {
        delete e;
        fprintf(stderr, "Image not found: \"%s\"\n", shpname);
        numlayers = 0;
        delete shpname;
        return;
    }
    delete[] shpname;
    shpnum = imgPool->size();
    imgPool->push_back(shpimage);
    shpnum <<= 16;
    for( i = 0; i < numlayers; i++ ) {
        /* get layer offsets from inifile */
        shpnums[i] = shpnum;
        shpnum += 32;
    }
    is_infantry = false;
    tmpnumshps = unitini->readInt(tname, "unittype",0);

    if( tmpnumshps == 0 )
        is_infantry = true;

    tmpspeed = unitini->readInt(tname, "speed");
    if (is_infantry) {
        if (tmpspeed == INIERROR) {
            speed = 4; // default for infantry is slower
            movemod = 1;
        } else {
            speed = (tmpspeed>4)?2:(7-tmpspeed);
            movemod = (tmpspeed>4)?(tmpspeed-4):1;
        }
    } else {
        if (tmpspeed == INIERROR) {
            speed = 2;
            movemod = 1;
        } else {
            speed = (tmpspeed>4)?2:(7-tmpspeed);
            movemod = (tmpspeed>4)?(tmpspeed-4):1;
        }
    }
    if (is_infantry) {
        talkmode = unitini->readString(tname, "talkback","Generic");
        sight = unitini->readInt(tname, "sight", 3);
    } else {
        talkmode = unitini->readString(tname, "talkback","Generic-Vehicle");
        sight = unitini->readInt(tname, "sight", 5);
    }
    createtalkback(talkmode);
    delete[] talkmode;
    maxhealth = unitini->readInt(tname, "health", 50);
    tmpspeed = unitini->readInt(tname, "turnspeed");
    if (tmpspeed == INIERROR) { // no default for infantry as do not turn
        turnspeed = 2;
        turnmod = 1;
    } else {
        turnspeed = (tmpspeed>4)?2:(7-tmpspeed);
        turnmod = (tmpspeed>4)?(tmpspeed-4):1;
    }
    if( is_infantry ) {
        //      size = 1;
        offset = 0;
    } else {
        //size = shpimage->getWidth();
        offset = (shpimage->getWidth()-24)>>1;
    }
    miscnames = unitini->readString(tname, "primary_weapon");
    if( miscnames == NULL ) {
        primary_weapon = NULL;
    } else {
        primary_weapon = wpool->getWeapon(miscnames);
        delete[] miscnames;
    }
    miscnames = unitini->readString(tname, "secondary_weapon");
    if( miscnames == NULL ) {
        secondary_weapon = NULL;
    } else {
        secondary_weapon = wpool->getWeapon(miscnames);
        delete[] miscnames;
    }
    deploytarget = unitini->readString(tname, "deploysto");
    if (deploytarget != NULL) {
        deployable = true;
        deploytype = uspool->getStructureTypeByName(deploytarget);
    } else {
        deploytype = NULL;
        maxpassengers = unitini->readInt(tname, "maxpassengers",0);
        if (maxpassengers > 0)
            deployable = true;
        else
            deployable = false;
    }
    miscnames = unitini->readString(tname,"armour");
    if (miscnames == NULL)
        armour = 0;
    else {
        if (strncasecmp(miscnames,"none",4) == 0)
            armour = 0;
        else if (strncasecmp(miscnames,"wood",4) == 0)
            armour = 1;
        else if (strncasecmp(miscnames,"light",5) == 0)
            armour = 2;
        else if (strncasecmp(miscnames,"heavy",5) == 0)
            armour = 3;
        else if (strncasecmp(miscnames,"concrete",8) == 0)
            armour = 4;
    }
}

char* UnitType::getRandTalk(Uint8 type) {
    Uint8 rnd,sze;
    std::vector<char*>* vecptr;
    switch (type) {
    case 0: /* report */
        vecptr = &tbReport;
        break;
    case 1: /* ack */
        vecptr = &tbAck;
        break;
    case 2: /* attack unit */
        vecptr = &tbAck;
        break;
    case 3: /* attack structure */
        vecptr = &tbAck;
        if (tbBAtk.size() > 0) {
            vecptr = &tbBAtk;
        }
        break;
    case 4: /* died */
        if (tbDie.size() > 0) {
            vecptr = &tbDie;
        } else {
            vecptr = NULL;
        }
        break;
    case 5: /* killed a unit */
        if (tbPKill.size() > 0) {
            vecptr = &tbPKill;
        } else {
            vecptr = NULL;
        }
        break;
    default:
        fprintf(stderr,"Unknown talkback type: %i\n",type);
        vecptr = NULL;
        break;
    }
    if (vecptr != NULL) {
        sze = (*vecptr).size();
        rnd = (int) ((double)sze*rand()/(RAND_MAX+1.0));
        return (*vecptr)[rnd];
    } else {
        return NULL;
    }
}

void UnitType::gettbvecptr(char** line, std::vector<char*>** vecptr) {
    if (strncasecmp((*line),"report",6) == 0) {
        (*vecptr) = &tbReport;
        sscanf((*line),"report=%s\n",(*line));
    } else if (strncasecmp((*line),"ack",3) == 0) {
        (*vecptr) = &tbAck;
        sscanf((*line),"ack=%s\n",(*line));
    } else if (strncasecmp((*line),"die",3) == 0) {
        (*vecptr) = &tbDie;
        sscanf((*line),"die=%s\n",(*line));
    } else if (strncasecmp((*line),"postkill",8) == 0) {
        (*vecptr) = &tbPKill;
        sscanf((*line),"postkill=%s\n",(*line));
    } else if (strncasecmp((*line),"buildatk",8) == 0) {
        (*vecptr) = &tbBAtk;
        sscanf((*line),"buildatk=%s\n",(*line));
    }
}

void UnitType::createtalkback(char* talkback) {
    INIFile* tbini;
    char* line;
    std::vector<char*>* vecptr;
    tbini = new INIFile("talkback.ini");
    if (tbini->seekSection(talkback) == 0) {
        fprintf(stderr,"Could not find talkback \"%s\", reverting to default\n",talkback);
        tbini->seekSection("Generic");
    }

    line = tbini->getLine();
    while ((line != NULL)&&(strncasecmp(line,"[",1))) {
        if ((strlen(line) > 7)&&(strncasecmp(line,"include",7) == 0)) {
            sscanf(line,"include=%s",line);
            if (strcasecmp(line,talkback) != 0) {
                createtalkback(line);
            } else {
                fprintf(stderr,"Warning, skipping self-referential include in %s\n",talkback);
            }
            delete[] line;
        } else {
            vecptr = NULL;
            if (strlen(line) < 5) {
                fprintf(stderr,"Line is blank? \"%s\"\n",line);
                vecptr = NULL;
            } else if (strncasecmp(line,"delete",6) == 0) {
                sscanf(line,"delete=%s\n",line);
                gettbvecptr(&line,&vecptr);
                (*vecptr).clear();
                vecptr = NULL;
            } else {
                gettbvecptr(&line,&vecptr);
            }
            if (vecptr != NULL) {
                (*vecptr).push_back(line);
            }
        }
        //delete[] line;
        line = tbini->getLine();
    }
    delete tbini;
}

UnitType::~UnitType() {
    Uint16 i;
    delete[] shpnums;
    for (i=0;i<tbReport.size();++i)
        delete[] tbReport[i];
    for (i=0;i<tbAck.size();++i)
        delete[] tbAck[i];
    for (i=0;i<tbDie.size();++i)
        delete[] tbDie[i];
    for (i=0;i<tbPKill.size();++i)
        delete[] tbPKill[i];
    for (i=0;i<tbBAtk.size();++i)
        delete[] tbBAtk[i];
}

/* note to self, pass owner, cellpos, facing and health to this
 (maybe subcellpos)*/
Unit::Unit(UnitType *type, Uint16 cellpos, Uint8 subpos, InfantryGroup *group, Uint8 owner) {
    int i;
    this->type = type;
    imagenumbers = new Uint16[type->getNumLayers()];
    for( i = 0; i < type->getNumLayers(); i++ ) {
        imagenumbers[i] = 0;
        if( owner != 0xff ) {
            imagenumbers[i] |= (uspool->getMap()->getPlayerPool()->getPaletteNum(owner)<<11);
        }
    }
    deleted = false;
    stopping = false;
    this->owner = owner;
    references = 0;
    this->cellpos = cellpos;
    this->subpos = subpos;
    xoffset = 0;
    yoffset = 0;
    health = type->getMaxHealth();
    selected = false;
    ratio = (double)health / (double)type->getMaxHealth();
    infgrp = group;

    if( infgrp != NULL )
        if( infgrp->isClear(subpos) ) /* else select another subpos */
            infgrp->addInfantry(this, subpos);

    PriAnim = NULL;
    SecAnim = NULL;
    TerAnim = NULL;
    deployed = false;
    passengers = 0;
    uspool->getMap()->getPlayerPool()->getPlayer(owner)->builtUnit(this);
}

Unit::~Unit() {
    delete[] imagenumbers;
    if( type->isInfantry() && infgrp != NULL ) {
        infgrp->removeInfantry(subpos);
        if( infgrp->getNumInfantry() == 0 )
            delete infgrp;
    }
    if (deployed) {
        (uspool->getsfxeng())->queueSound("constru2.aud");
        (uspool->getsfxeng())->queueSound("hvydoor1.aud");
        uspool->createStructure(type->getDeployTarget(),calcDeployPos(),owner,true);
    }
}

void Unit::remove() {
    uspool->getMap()->getPlayerPool()->getPlayer(owner)->lostUnit(this,deployed);
    deleted = true;
    if(references == 0) {
        delete this;
    } else {
        //fprintf(stderr,"Unit::remove: did not delete due to unresolved references. (%p,%i)\n",this,references);
    }
}


Uint8 Unit::getImageNums(Uint32 **inums, Sint8 **xoffsets, Sint8 **yoffsets) {
    int i;

    Uint32 *shpnums;

    shpnums = type->getSHPNums();

    *inums = new Uint32[type->getNumLayers()];
    *xoffsets = new Sint8[type->getNumLayers()];
    *yoffsets = new Sint8[type->getNumLayers()];
    for(i = 0; i < type->getNumLayers(); i++ ) {
        (*inums)[i] = shpnums[i]+imagenumbers[i];
        (*xoffsets)[i] = xoffset-type->getOffset();
        (*yoffsets)[i] = yoffset-type->getOffset();
    }

    return type->getNumLayers();
}

void Unit::move(Uint16 dest) {
    MoveAnimEvent* nextmove;
    if (!stopping) {
        nextmove = new MoveAnimEvent(type->getSpeed(), this,dest);
        setPriAnim(nextmove,true);
    } else {
        //fprintf(stderr,"Unit::move: was stopping, not moving!\n");
    }
}

void Unit::attackUnit(Unit *un) {
    if (!stopping) {
        setPriAnim(new UAttackAnimEvent(0, this, un),true);
    }
}

void Unit::attackStructure(Structure *st) {
    if (!stopping) {
        setPriAnim(new UAttackAnimEvent(0, this, st),true);
    }
}

void Unit::attackUnitOrStructure(UnitOrStructure* uos) {
    if (!stopping) {
        setPriAnim(new UAttackAnimEvent(0, this, uos),true);
    }
}

void Unit::stop() {}

void Unit::unrefer() {
    references--;
    if( deleted && references == 0 ) {
        //fprintf(stderr,"all references resolved, deleting. (%p)\n",this);
        delete this;
    }
}

void Unit::applyDamage(Sint16 amount, Weapon* weap) {
    //fprintf(stderr,"%i * %f = ",amount,weap->getVersus(type->getArmour()));
    amount = (Sint16)((double)amount * weap->getVersus(type->getArmour()));
    //fprintf(stderr,"%i (a == %i)\n",amount,type->getArmour());
    if ((health-amount) <= 0) {
        doRandTalk(4);
        // todo: add infantry death animation
        if (uspool->getMap()->getPlayerPool()->getPlayer(owner)->getAI() != NULL) {
            uspool->getMap()->getPlayerPool()->getPlayer(owner)->getAI()->attackOtherSide();
        }
        uspool->removeUnit(this);
        return;
    } else if ((health-amount) > type->getMaxHealth()) {
        health = type->getMaxHealth();
    } else {
        health -= amount;
    }
    ratio = (double)health / (double)type->getMaxHealth();
}

void Unit::doRandTalk(Uint8 ttype) {
    char* sname;
    sname = type->getRandTalk(ttype);
    if (sname != NULL) {
        (uspool->getsfxeng())->queueSound(sname);
    }
}

void Unit::setPriAnim(UnitAnimEvent* e, bool stop) {
    setAnim(e,&PriAnim,stop);
}

void Unit::setSecAnim(UnitAnimEvent* e, bool stop) {
    setAnim(e,&SecAnim,stop);
}

void Unit::setTerAnim(UnitAnimEvent* e, bool stop) {
    setAnim(e,&TerAnim,stop);
}

void Unit::setAnim(UnitAnimEvent* e, UnitAnimEvent** anim, bool stop) {
    if ((!stop)||((*anim) == NULL)) {
        (*anim) = e;
        uspool->scheduleUSEvent((*anim));
    } else {
        stopping = true;
        (*anim)->setSchedule(e);
        (*anim)->stop();
    }
}

bool Unit::canDeploy() {
    if (type->canDeploy()) {
        if (type->getDeployTarget() != NULL) {
            if (!deployed)
                return checkDeployTarget(calcDeployPos());
            return false;
        }
    } else {
        return (passengers > 0);
    }
    return false;
}

void Unit::deploy(void) {
    if (canDeploy()) { // error catching
        if (type->getDeployTarget() != NULL) {
            deployed = true;
            uspool->removeUnit(this);
        }
    }
}

bool Unit::checkDeployTarget(Uint32 pos) {
    Sint8 placexpos, placeypos;
    Uint32 mapwidth = uspool->getMap()->getWidth();
    Uint8 typewidth = type->getDeployType()->getXsize();
    Uint8 typeheight = type->getDeployType()->getYsize();
    if (pos == (Uint32)(-1)) {
        return false;
    }
    for( placeypos = -(typeheight>>1); placeypos < (typeheight>>1)+1; ++placeypos) {

        for( placexpos = -(typewidth>>1); placexpos < (typewidth>>1)+1; ++placexpos) {
            if( type->getDeployType()->isBlocked(placeypos*typewidth+placexpos) ) {
                if( uspool->getMap()->getCost(pos+placeypos*mapwidth+placexpos,this)>1 ) {
                    return false;
                }
            }
        }
    }
    return true;
}

Uint32 Unit::calcDeployPos() {
    Uint32 deploypos;
    Uint32 mapwidth = uspool->getMap()->getWidth();

    deploypos = cellpos;
    if ((Uint32)((type->getDeployType()->getXsize()) >> 1) > deploypos)
        return (Uint32)-1; // large number
    else
        deploypos -= (type->getDeployType()->getXsize()) >> 1;
    if ((mapwidth*((type->getDeployType()->getYsize()) >> 1)) > deploypos)
        return (Uint32)-1;
    else
        deploypos -= mapwidth*((type->getDeployType()->getYsize()) >> 1);
    return deploypos;
}
