#include <math.h>
#include "unitanimations.h"
#include "weaponspool.h"
#include "projectileanim.h"
#include "memtrack.h"
#include "inifile.h"

UnitAnimEvent::UnitAnimEvent(Uint32 p, Unit* un, Uint8 smode) : ActionEvent(p) {
    //fprintf(stderr,"debug: UnitAnim constructor %i (%p)\n",smode,this);
    this->un = un;
    un->referTo();
    scheduled = NULL;
    mode = smode;
    if (un->stopping) {
        //fprintf(stderr,"UAE::UAE unit is stopping\n");
    }
}

UnitAnimEvent::~UnitAnimEvent() {
    //fprintf(stderr,"debug: UnitAnim destructor schedmode: %i mymode: %i (%p)\n",
    //  (scheduled!=NULL)?scheduled->getMode():-1,this->getMode(),this);
    if (scheduled != NULL) {
        switch (scheduled->getMode()) {
        case 0:
            un->setPriAnim(scheduled,false);
            break;
        case 1:
            un->setSecAnim(scheduled,false);
            break;
        case 2:
            un->setTerAnim(scheduled,false);
            break;
        default:
            fprintf(stderr,"WARNING INVALID MODE ON UNITANIMEVENT (%i)\n",mode);
            abort();
            break;
        }
    } else {
        //fprintf(stderr,"Debug: setting something to NULL (%i)\n",this->getMode());
        switch (this->getMode()) {
        case 0:
            un->PriAnim = NULL;
            break;
        case 1:
            un->SecAnim = NULL;
            break;
        case 2:
            un->TerAnim = NULL;
            break;
        default:
            fprintf(stderr,"WARNING: INVALID MODE ON UNITANIMEVENT (%i)\n",mode);
            abort();
            break;
        }
    }
    //fprintf(stderr,"P%i S%i T%i\n",(NULL==un->PriAnim),
    //        (NULL==un->SecAnim),(NULL==un->TerAnim));
    un->stopping = false;
    un->unrefer();
}

void UnitAnimEvent::setSchedule(UnitAnimEvent* e) {
    //fprintf(stderr,"debug: Scheduling an event. (this: %p, e: %p)\n",this,e);
    if (scheduled != NULL) {
        scheduled->setSchedule(NULL);
        scheduled->stop();
    }
    scheduled = e;
}

// These functions are specific to an animation type

void UnitAnimEvent::setRange(Uint32 nr) {}

void UnitAnimEvent::changedir(Uint8 dir) {}

u_animinfo_t UnitAnimEvent::parseAnimEntry(char* name) {
    u_animinfo_t retval;
    INIFile* animini = new INIFile("animations.ini");
    animini->seekSection(name);
    delete animini;
    return retval;
}

/*
GenericAnimEvent::GenericAnimEvent(Uint32 p, Unit* un, ...) : UnitAnimEvent(p,un,0)
{
     stopping = false;
     mode = 0;
     animinfo = parseAnimEntry("foo");
     un->uspool->scheduleUSEvent(this);
}
*/

MoveAnimEvent::MoveAnimEvent(Uint32 p, Unit* un, Uint16 dest) : UnitAnimEvent(p,un,0) {
    //fprintf(stderr,"debug: MoveAnim constructor (this %p un %p dest %u)\n",this,un,dest);
    stopping = false;
    mode = 0;
    blocked = false;
    walking = false;
    range = 0;
    this->dest = dest;
    this->un = un;
    path = NULL;
    newpos = 0xffff;
    istep = 0;
}

MoveAnimEvent::~MoveAnimEvent() {
    //fprintf(stderr,"debug: MoveAnim destructor (this %p un %p dest %u)\n",this,un,dest);
    if (path != NULL) {
        delete path;
    }
    if ((walking)&&(un->SecAnim != NULL)) {
        un->SecAnim->stop();
    }
}

void MoveAnimEvent::run() {
    Sint8 uxoff, uyoff;
    Uint8 oldsubpos;

    if( !un->isAlive() ) {
        delete this;
        return;
    }

    if( path == NULL ) {
        (un->getUSPool())->setCostCalcOwnerAndType(un->owner, 0);
        path = new Path(un->getPos(), dest, (un->getUSPool())->getMap(), range);
        if( !path->empty() ) {
            startMoveOne(false);
        } else {
            delete this;
        }
        return;
    }

    if( blocked ) {
        blocked = false;
        startMoveOne(true);
        return;
    }
    /* if distance left is smaller than xmod we're ready */

    un->xoffset += xmod;
    un->yoffset += ymod;

    if( !moved_half && (abs(un->xoffset) >= 12 || abs(un->yoffset) >= 12) ) {
        oldsubpos = un->subpos;
        un->subpos = (un->getUSPool())->postMove(un, newpos);
        un->cellpos = newpos;
        un->xoffset = -un->xoffset;
        un->yoffset = -un->yoffset;
        if( un->type->isInfantry() ) {
            un->infgrp->getSubposOffsets(oldsubpos, un->subpos, uxoff, uyoff);
            un->xoffset += uxoff;
            un->yoffset += uyoff;
            xmod = 0;
            ymod = 0;
            if( un->xoffset < 0 )
                xmod = 1;
            else if( un->xoffset > 0 )
                xmod = -1;
            if( un->yoffset < 0 )
                ymod = 1;
            else if( un->yoffset > 0 )
                ymod = -1;
            //     printf("%d, %d - moving %d, %d\n", un->xoffset, un->yoffset, xmod, ymod);
        }
        moved_half = true;
    }

    if( abs(un->xoffset) < abs(xmod) )
        xmod = 0;
    if( abs(un->yoffset) < abs(ymod) )
        ymod = 0;

    if( xmod == 0 && ymod == 0 ) {
        un->xoffset = 0;
        un->yoffset = 0;
        moveDone();
        return;
    }

    (un->getUSPool())->scheduleUSEvent(this);
}

void MoveAnimEvent::startMoveOne(bool wasblocked) {
    Uint8 face;

    newpos = (un->getUSPool())->preMove(un, path->top(), &xmod, &ymod);
    if( newpos == 0xffff ) {
        delete path;
        path = NULL;
        (un->getUSPool())->setCostCalcOwnerAndType(un->owner, 0);
        path = new Path(un->getPos(), dest, (un->getUSPool())->getMap(), range);
        if( path->empty() ) {
            xmod = 0;
            ymod = 0;
            (un->getUSPool())->scheduleUSEvent(this);
            return;
        }
        newpos = (un->getUSPool())->preMove(un, path->top(), &xmod, &ymod);
        if( newpos == 0xffff ) {
            if (wasblocked) {
                xmod = 0;
                ymod = 0;
                this->stop();
                (un->getUSPool())->scheduleUSEvent(this);
                return;
            } else {
                /* TODO: tell the blocking unit to move here */

                blocked = true;

                if (walking) {
                    un->SecAnim->stop();
                    walking = false;
                }

                (un->getUSPool())->scheduleUSEvent(this);
                return;
            }
        }
    }

    face = (32-(path->top() << 2))&0x1f;
    path->pop();

    moved_half = false;

    if (((UnitType *)un->getType())->isInfantry()) {
        if (walking) {
            un->SecAnim->changedir(face);
        } else {
            un->setSecAnim(new WalkAnimEvent(((UnitType *)un->getType())->getSpeed(), un, face, 0),true);
            walking = true;
        }
        (un->getUSPool())->scheduleUSEvent(this);

    } else {
        Uint8 curface = (un->getImageNum(0)&0x1f);
        Uint8 delta = (abs(curface-face))&0x1f;
        if( (un->getImageNum(0)&0x1f) != face ) {
            if( (delta < 5) || (delta > 27) ) {
                un->setSecAnim(new TurnAnimEvent(((UnitType *)un->getType())->getROT(), un, face, 0, NULL,1),false);
                (un->getUSPool())->scheduleUSEvent(this);
            } else {
                un->setPriAnim(new TurnAnimEvent(((UnitType *)un->getType())->getROT(), un, face, 0, this,0),false);
                return;
            }
        } else {
            (un->getUSPool())->scheduleUSEvent(this);
        }
    }
}

void MoveAnimEvent::moveDone() {
    un->xoffset = 0;
    un->yoffset = 0;

    if( !path->empty() && !stopping ) {
        startMoveOne(false);
    } else {
        if( dest != un->getPos() && !stopping ) {
            delete path;
            path = NULL;
            (un->getUSPool())->setCostCalcOwnerAndType(un->owner, 0);
            path = new Path( un->getPos(), dest, (un->getUSPool())->getMap(), range );
        }
        if( path->empty() || stopping ) {
            delete this;
        } else {
            startMoveOne(false);
        }
    }
}

void MoveAnimEvent::stop() {
    un->getUSPool()->abortMove(un,newpos);
    stopping = true;
}

WalkAnimEvent::WalkAnimEvent(Uint32 p, Unit *un, Uint8 dir, Uint8 layer) : UnitAnimEvent(p,un,1) {
    //fprintf(stderr,"debug: WalkAnim constructor\n");
    this->un = un;
    this->dir = dir;
    this->layer = layer;
    stopping = false;
    istep = 0;
    calcbaseimage();
    mode = 1;
}

WalkAnimEvent::~WalkAnimEvent() {
    //fprintf(stderr,"debug: WalkAnim destructor\n");
}

void WalkAnimEvent::run() {
    Uint8 layerface;
    if (!stopping) {
        layerface = baseimage + istep;
        istep = (istep + 1)%6;
        un->setIFace(layer, layerface);
        (un->getUSPool())->scheduleUSEvent(this);
    } else {
        un->setIFace(layer,dir>>2);
        delete this;
        return;
    }
}

void WalkAnimEvent::calcbaseimage() {
    baseimage = 16 + 3*(dir/2);
}

TurnAnimEvent::TurnAnimEvent(Uint32 p, Unit *un, Uint8 dir, Uint8 layer,
                             UnitAnimEvent *e, Uint8 smode) : UnitAnimEvent(p,un,smode) {
    //fprintf(stderr,"TurnAnim: constructor\n");
    Uint8 layerface;
    this->un = un;
    this->dir = dir;
    this->layer = layer;
    runonce = false;
    if (e != NULL) {
        setSchedule(e);
    }
    mode = smode;
    stopping = false;
    layerface = un->getImageNum(layer)&0x1f;
    if( ((layerface-dir)&0x1f) < ((dir-layerface)&0x1f) ) {
        turnmod = -(((UnitType *)un->getType())->getTurnMod());
    } else {
        turnmod = (((UnitType *)un->getType())->getTurnMod());
    }
}

TurnAnimEvent::~TurnAnimEvent() {
    //fprintf(stderr,"TurnAnim: destructor (%p)\n",this);
}

void TurnAnimEvent::run() {
    Uint8 layerface;

    if (stopping) {
        //fprintf(stderr,"TurnAnim: stopping\n");
        delete this;
        return;
    }

    if (!runonce) {
        runonce = true;
        if (un->getType()->getNumLayers() > 1) {
            //fprintf(stderr,"TurnAnimEvent: Scheduling a TurretAnimEvent (%p)\n",this);
            un->setTerAnim(new TurretAnimEvent(un->type->getROT(), un, dir, 1, NULL, 2),false);
        }
    }
    layerface = un->getImageNum(layer)&0x1f;
    if( abs((layerface-dir)&0x1f) > abs(turnmod) ) {
        layerface += turnmod;
        layerface &= 0x1f;
    } else
        layerface = dir;

    un->setFace(layer, layerface);
    if( layerface == dir || !un->isAlive()) {
        //fprintf(stderr,"TurnAnim: done\n");
        delete this;
        return;
    }
    (un->getUSPool())->scheduleUSEvent(this);
}

TurretAnimEvent::TurretAnimEvent(Uint32 p, Unit* un, Uint8 dir, Uint8 layer, UnitAnimEvent* e, Uint8 smode) :
UnitAnimEvent(p,un,smode) {
    //fprintf(stderr,"TurretAnim: constructor %i (%p)\n",smode,this);
    Uint8 layerface;
    this->un = un;
    this->dir = dir;
    this->layer = layer;

    if (e != NULL) {
        //fprintf(stderr,"TurretAnim: scheduling an event\n");
        setSchedule(e);
    }
    mode = smode;

    stopping = false;

    layerface = un->getImageNum(layer)&0x1f;
    if( dir == layerface ) {
        //fprintf(stderr,"\tTurretAnim: done (in constructor)\n");
        stopping = true;
        return;
    }
    if( ((layerface-dir)&0x1f) < ((dir-layerface)&0x1f) ) {
        turnmod = -1;
    } else {
        turnmod = 1;
    }
}

TurretAnimEvent::~TurretAnimEvent() {
    //fprintf(stderr,"TurretAnim: destructor\n");
}

void TurretAnimEvent::run() {
    Uint8 layerface;

    if (stopping) {
        //fprintf(stderr,"TurretAnim: stopping (in run) (%p)\n",this);
        delete this;
        return;
    }

    layerface = un->getImageNum(layer)&0x1f;
    if( abs((layerface-dir)&0x1f) > abs(turnmod) ) {
        layerface += turnmod;
        layerface &= 0x1f;
    } else {
        layerface = dir;
    }

    un->setFace(layer, layerface);
    if( layerface == dir) {
        //fprintf(stderr,"TurretAnim: done (in run) (%p)\n",this);
        delete this;
        return;
    }
    (un->getUSPool())->scheduleUSEvent(this);
}

UAttackAnimEvent::UAttackAnimEvent(Uint32 p, Unit *un, UnitOrStructure *target) : UnitAnimEvent(p,un,0) {
    //fprintf(stderr,"debug: UAttackAnim destructor\n");
    this->un = un;
    this->target = target;
    stopping = false;
    target->referTo();
    mode = 0;
}

UAttackAnimEvent::~UAttackAnimEvent() {
    //fprintf(stderr,"debug: UAttackAnim destructor\n");
    target->unrefer();
}

void UAttackAnimEvent::stop() {
    if (un == NULL) {
        fprintf(stderr,"UAttackAnimEvent::stop: un is NULL!?\n");
        abort();
    }
    stopping = true;
}

void UAttackAnimEvent::run() {
    Uint32 distance;
    Sint32 xtiles, ytiles;
    Uint16 atkpos;
    float alpha;
    Uint8 facing;

    if( !un->isAlive() || stopping ) {
        delete this;
        return;
    }

    if( !target->isAlive() || stopping) {
        if ( !target->isAlive() ) {
            if (un->getUSPool()->getMap()->getPlayerPool()->getPlayer(un->getOwner())->getAI() != NULL) {
                un->getUSPool()->getMap()->getPlayerPool()->getPlayer(un->getOwner())->getAI()->attackOtherSide();
            }
            un->doRandTalk(5);
        }
        delete this;
        return;
    }
    atkpos = target->getPos();

    xtiles = un->cellpos % (un->getUSPool())->getMap()->getWidth() - atkpos % (un->getUSPool())->getMap()->getWidth();
    ytiles = un->cellpos / (un->getUSPool())->getMap()->getWidth() - atkpos / (un->getUSPool())->getMap()->getWidth();
    distance = abs(xtiles)>abs(ytiles)?abs(xtiles):abs(ytiles);

    if( distance > un->type->getWeapon()->getRange() /* weapons range */ ) {
        setDelay(0);
        //fprintf(stderr,"UAttackAnimEvent::run: Scheduling a MoveAnimEvent\n");
        MoveAnimEvent* nme = new MoveAnimEvent(un->type->getSpeed(), un, atkpos);
        nme->setRange(un->type->getWeapon()->getRange());
        nme->setSchedule(this);
        un->setPriAnim(nme,false);
        return;
    }
    //Make sure we're facing the right way
    if( xtiles == 0 ) {
        if( ytiles < 0 ) {
            alpha = -3.1415926536f/2.0f;
        } else {
            alpha = 3.1415926536f/2.0f;
        }
    } else {
        alpha = atan((float)ytiles/(float)xtiles);
        if( xtiles < 0 ) {
            alpha = 3.1415926536f+alpha;
        }
    }
    facing = (40-(Sint8)(alpha*16/3.1415926536f))&0x1f;
    if (un->type->isInfantry()) {
        if (facing != un->getImageNum(0)&0x1f) {
            un->setIFace(0,facing>>2);
        }
    } else if (un->type->getNumLayers() > 1 ) {
        if (abs(facing - un->getImageNum(1)&0x1f) > un->type->getROT()) {
            setDelay(0);
            //fprintf(stderr,"UAttackAnimEvent::run: Scheduling a TurretAnimEvent\n");
            un->setPriAnim(new TurretAnimEvent(un->type->getROT(), un, facing, 1, this, 0),false);
            return;
        }
    } else {
        if (abs(facing - un->getImageNum(0)&0x1f) > un->type->getROT()) {
            setDelay(0);
            un->setPriAnim(new TurnAnimEvent(((UnitType *)un->getType())->getROT(), un, facing, 0, this,0),false);
            return;
        }
    }
    // We can shoot

    un->type->getWeapon()->fire(un, target->getBPos(), target->getSubpos(), (un->getUSPool()) );
    // set delay to reloadtime
    setDelay(un->type->getWeapon()->getReloadTime());
    (un->getUSPool())->scheduleUSEvent(this);
}

