#include "weaponspool.h"
#include "inifile.h"
//#include "sound.h"
#include "shpimage.h"
#include "unit.h"
#include "structure.h"
#include "unitandstructurepool.h"
#include "projectileanim.h"
#include "game.h"
#include "unitorstructure.h"
#include "memtrack.h"

using namespace std;

Warhead::Warhead(char *whname, INIFile *weapini, WeaponsPool *wp) {
    char *tmpname;
    SHPImage* temp;

    tmpname = weapini->readString(whname, "explosionimage");
    explosionimage = 0;
    if( tmpname != NULL ) {
        explosionimage = wp->imgpool->size()<<16;
        temp = new SHPImage(tmpname, Game::mapscaleq);
        wp->imgpool->push_back(temp);
        delete[] tmpname;
    }
    explosionanimsteps = temp->getNumImg();
    //  weapini->readInt(whname, "explosionanimsteps", 1);
    //	printf("Number of images in %s's explosion: %i\n",whname,explosionanimsteps);
    explosionsound = weapini->readString(whname, "explosionsound");
    infantrydeath = 0;
    walls = (bool)weapini->readInt(whname, "walls",0);
    trees = false;
    //blastradius = 0;
    tmpname = weapini->readString(whname, "versus");
    versus[0] = 100;
    versus[1] = 100;
    versus[2] = 100;
    versus[3] = 100;
    versus[4] = 100;
    if (tmpname != NULL) {
        sscanf(tmpname,"%u,%u,%u,%u,%u",&versus[0],&versus[1],
               &versus[2],&versus[3],&versus[4]);
    }
    //for (int i=0;i<5;++i)
    //    fprintf(stderr,"%s\t%i\t%i\n",whname,i,versus[i]);
}

Warhead::~Warhead() {
    if( explosionsound != NULL ) {
        delete[] explosionsound;
    }
}

Projectile::Projectile(char *pname, INIFile *weapini, WeaponsPool *wp) {
    char *iname = weapini->readString(pname, "image");
    SHPImage* temp;
    imagenum = 0;
    rotates = false;
    rotationimgs = 0;
    if( iname != NULL ) {
        imagenum = wp->imgpool->size()<<16;
        temp = new SHPImage(iname, Game::mapscaleq);
        //		printf("Projectile %s has %s which has %i\n",pname,iname,temp->getNumImg());
        wp->imgpool->push_back(temp);
        delete[] iname;

        if ((bool)weapini->readInt(pname,"rotates",0)) {
            rotationimgs = temp->getNumImg();
            rotates = true;
        }
    }
    AA = false;
    AG = true;
    high = false;
    inacurate = false;
}

Projectile::~Projectile() {}

Weapon::Weapon(char *wname, WeaponsPool *wp) {
    char *pname, *whname, *faname, *faimage;
    map<string, Projectile*>::iterator projentry;
    map<string, Warhead*>::iterator wheadentry;
    INIFile *weapini = wp->getWeaponsINI();
    SHPImage* fireanimtemp;
    Uint8 additional, i;

    this->wp = wp;
    pname = weapini->readString(wname, "projectile");
    if( pname == NULL ) {
        fprintf(stderr, "Unable to find projectile for weapon \"%s\" in inifile..\n", wname);
        throw 0;
    }
    projentry = wp->projectilepool.find((string)pname);
    if( projentry == wp->projectilepool.end() ) {
        try {
            projectile = new Projectile(pname, weapini, wp);
        } catch( int e ) {
            fprintf(stderr, "Unable to find projectile \"%s\" used for weapon \"%s\".\nUnit using this weapon will be unarmed\n", pname, wname);
            delete[] pname;
            throw;
        }
        wp->projectilepool[(string)pname] = projectile;
    } else {
        projectile = projentry->second;
    }
    delete[] pname;

    whname = weapini->readString(wname, "warhead");
    if( whname == NULL ) {
        fprintf(stderr, "Unable to find warhead for weapon \"%s\" in inifile..\n", wname);
        throw 0;
    }
    wheadentry = wp->warheadpool.find((string)whname);
    if( wheadentry == wp->warheadpool.end() ) {
        try {
            whead = new Warhead(whname, weapini, wp);
        } catch( int e ) {
            fprintf(stderr, "Unable to find Warhead \"%s\" used for weapon \"%s\".\nUnit using this weapon will be unarmed\n", whname, wname);
            delete[] whname;
            throw;
        }
        wp->warheadpool[(string)whname] = whead;
    } else {
        whead = wheadentry->second;
    }
    delete[] whname;

    speed = weapini->readInt(wname, "speed", 100);
    range = weapini->readInt(wname, "range", 1);
    reloadtime = weapini->readInt(wname, "reloadtime", 5);
    damage = weapini->readInt(wname, "damage", 10);
    burst = weapini->readInt(wname, "burst", 1);
    heatseek = (bool)weapini->readInt(wname, "heatseek", 0);
    inaccurate = (bool)weapini->readInt(wname, "inaccurate", 0);
    fireimage = wp->imgpool->size()<<16;
    //    wp->imgpool->push_back(new SHPImage("minigun.shp", Game::mapscaleq));
    firesound = weapini->readString(wname, "firesound");

    faname = weapini->readString(wname, "fireimage", "none");
    if (strcasecmp(faname,"none") == 0) {
        delete[] faname;
        numfireimages = 0;
        numfiredirections = 1;
        fireimage = 0;
    } else {
        additional = (Uint8)weapini->readInt(faname,"additional",0);
        faimage = weapini->readString(faname, "image", "minigun.shp");
        fireanimtemp = new SHPImage(faimage, Game::mapscaleq);
        numfireimages = fireanimtemp->getNumImg();
        numfiredirections = weapini->readInt(faname, "directions", 1);
        if (numfiredirections == 0) {
            numfiredirections = 1;
        }
        fireimages = new Uint32[numfiredirections];
        fireimages[0] = fireimage;
        wp->imgpool->push_back(fireanimtemp);
        if (additional != 0) {
            char* tmpname = new char[12];
            for (i=2;i<=additional;++i) {
                sprintf(tmpname,"image%i",i);
                faimage = weapini->readString(faname, tmpname, "");
                if (strcasecmp(faimage,"") != 0) {
                    fireanimtemp = new SHPImage(faimage, Game::mapscaleq);
                    fireimages[i-1]=(wp->imgpool->size()<<16);
                    numfireimages += fireanimtemp->getNumImg();
                    wp->imgpool->push_back(fireanimtemp);
                } else {
                    fireimages[i] = 0;
                    fprintf(stderr,"Warning: %s was empty in [%s]\n",tmpname,faname);
                }
            }
            delete[] tmpname;
        } else if (numfiredirections != 1) {
            for (i=1;i<numfiredirections;++i) {
                fireimages[i] = fireimage+i*(numfireimages/numfiredirections);
            }
        }
        if (faname != NULL) {
            delete[] faname;
        }
        if (faimage != NULL) {
            delete[] faimage;
        }
    }
}

Weapon::~Weapon() {
    if( firesound != NULL ) {
        delete[] firesound;
    }
    if (fireimage != 0) {
        delete[] fireimages;
    }
}

void Weapon::fire(UnitOrStructure *owner, Uint16 target, Uint8 subtarget, UnitAndStructurePool *uspool) {
    if( firesound != NULL ) {
        uspool->getsfxeng()->queueSound(firesound);
    }
    if( fireimage != 0 ) {
        Uint32 length = numfireimages;
        Uint8 facing;
        if (owner->getType()->getNumLayers() == 1) {
            facing = (owner->getImageNum(0))&0x1f;
        } else {
            facing = (owner->getImageNum(1))&0x1f;
        }
        if (!owner->getType()->isInfantry()) {
            facing >>= 2;
        }
        length /= numfiredirections;
        if (numfiredirections == 1) {
            facing = 0;
        }
        new ExplosionAnim(1, uspool, owner->getPos(),fireimages[facing],
                          (Uint8)length,/*owner->getXoffset()+*/InfantryGroup::getunitoffsets()[owner->getSubpos()],
                          /*owner->getYoffset()+*/InfantryGroup::getunitoffsets()[owner->getSubpos()]);
    }
    uspool->scheduleUSEvent(new ProjectileAnim(0, this, owner, target, subtarget, uspool));
}

WeaponsPool::WeaponsPool(vector<SHPImage*> *imgpool) {
    this->imgpool = imgpool;
    weapini = new INIFile("weapons.ini");
}

WeaponsPool::~WeaponsPool() {
    /* FIXME: go through the maps and free all memory */
    delete weapini;
    weaponspool.clear();
    projectilepool.clear();
    warheadpool.clear();
}

Weapon *WeaponsPool::getWeapon(char *wname) {
    map<string, Weapon*>::iterator weapentry;
    Weapon *weap;

    weapentry = weaponspool.find((string)wname);

    if( weapentry == weaponspool.end() ) {
        try {
            weap = new Weapon(wname, this);
        } catch( int e ) {
            /* if weapon is NULL unit doesn't have a weapon */
            return NULL;
        }
        weaponspool[(string)wname] = weap;
        return weap;
    } else {
        return weapentry->second;
    }
}
