/*****************************************************************************
 * input.h - header file for input
 *
 ****************************************************************************/

#ifndef __INPUT_H
#define __INPUT_H

#include "SDL.h"
#include "map.h"
#include "sidebar.h"
#include "selection.h"
#include "cursor.h"

#define KBDMOD_SHIFT 1
#define KBDMOD_CTRL 2
#define KBDMOD_ALT 4

#define MDOWN_NONE 0
#define MDOWN_MAP 1
#define MDOWN_SIDEBAR 2
#define MDOWN_TAB 3

class SoundEngine;

class Input {
public:
    Input(Sidebar *sbar, CnCMap *map,
          Uint16 screenwidth, Uint16 screenheight,
          SDL_Rect *maparea, Cursor *cur, SoundEngine *sndengine);
    ~Input();
    void handle();
    Uint8 shouldQuit() {
        return done;
    }
    static bool isMinimapEnabled() {
        return minimapEnabled;
    }
    static bool isDrawing() {
        return drawing;
    }
    static SDL_Rect getMarkRect() {
        return markrect;
    }
private:
    void updateMousePos();
    void clickMap(int mx, int my);
    void clickedTile(int mx, int my, Uint16 &pos, Uint8 &subpos);
    void clickSidebar(int mx, int my);
    void setCursorNum(int mx, int my);
    void selectRegion();
    Uint16 checkPlace(int mx, int my);

    Uint8 done, donecount;
    Uint16 width;
    Uint16 height;
    CnCMap *map;
    Sidebar *sidebar;
    SDL_Rect *maparea;
    Cursor *cursor;
    Uint16 tabwidth;
    Uint8 tilewidth;
    Selection *selected;
    Uint8 kbdmod, lmousedown, rmousedown;

    SoundEngine *sndengine;

    static bool drawing;
    static SDL_Rect markrect;
    static bool minimapEnabled;

    bool placing_struct, deploying_super;
    StructureType *placetype;
    char placename[13];
    bool placeposvalid;
    bool temporary_place_unit;
};
#endif
