/*****************************************************************************
 * map.h - header file for map class
 *
 ****************************************************************************/

#ifndef __MAP_H
#define __MAP_H


#ifdef _MSC_VER
#pragma warning (disable: 4503)
#pragma warning (disable: 4786)
#endif


#include <vector>
#include <map>

class CnCMap;
class Unit;

#include "inifile.h"
#include "actioneventqueue.h"
#include "shpimage.h"
#include "unitandstructurepool.h"
#include "playerpool.h"

#define HAS_OVERLAY 0x100
#define HAS_TERRAIN 0x200

class TerrainEntry {
public:
    Uint32 shpnum;
    Sint16 xoffset;
    Sint16 yoffset;
};

typedef struct {
    Uint16 templateNum;
    Uint8 tileNum;
}
TileList;

typedef struct {
    Sint8 x, y;
    Uint8 t;
}
ScrollVector;

class CnCMap {
public:
    CnCMap( ActionEventQueue *evQueue, char *mapname, Uint8 maptype );
    ~CnCMap();

    SDL_Surface *getMapTile( Uint32 pos ) {
        return tileimages[tilematrix[pos]];
    }
    Uint32 getTiberium(Uint32 pos) {
        return (overlaymatrix[pos] & 0xF);
    }
    Uint32 getSmudge(Uint32 pos) {
        return ((overlaymatrix[pos] & 0xF0) << 12);
    }
    Uint32 setSmudge(Uint32 pos, Uint8 value) {
        /// clear the existing smudge bits first
        overlaymatrix[pos] &= ~(0xF0);
        return (overlaymatrix[pos] |= (value<<4));
    }
    Uint32 setTiberium(Uint32 pos, Uint8 value) {
        /// clear the existing tiberium bits first
        overlaymatrix[pos] &= ~0xF;
        return (overlaymatrix[pos] |= value);
    }
    Uint32 getOverlay(Uint32 pos);
    Uint32 getTerrain(Uint32 pos, Sint16 &xoff, Sint16 &yoff);


    std::vector<SHPImage *> *getImagePool();

    void scroll(Uint8 direction);
    void doscroll();
    void setMaxScroll(Uint32 x, Uint32 y);

    char *getTheme() {
        return theme;
    }
    char *getBrief() {
        return brief;
    }
    char *getAction() {
        return action;
    }
    char *getTheater() {
        return theater;
    }
    char *getWinMov() {
        return winmov;
    }
    char *getLoseMov() {
        return losemov;
    }

    UnitAndStructurePool *getUASP() {
        return uspool;
    }

    Uint32 getScrollPos() {
        return ypos*width+xpos;
    }
    Uint16 getXScroll() {
        return xpos;
    }
    Uint16 getYScroll() {
        return ypos;
    }
    Uint16 getWidth() {
        return width;
    }
    Uint16 getHeight() {
        return height;
    }
    //   char *getTheater(){return theater;}

    Uint16 getCost(Uint16 pos);
    Uint16 getCost(Uint16 pos,Unit* excpUn);
    //    Uint8 getSide(void);
    PlayerPool *getPlayerPool() {
        return &ppool;
    }
    SDL_Surface* getMiniMap();
    bool toScroll() {
        return toscroll;
    }
    void printPlayersUnitsStructures(); // only for development purposes
private:

    /* private functions used to load the map, these are in loadmap.cpp */

    void loadIni();     // load the ini part of the map
    void simpleSections(INIFile *inifile);  // the map section of the ini
    void advancedSections(INIFile *inifile); // the advanced section of the ini
    void loadBin();     // load the bin part of the map
    void unMapPack( INIFile* inifile);   // extract ra map data
    void loadPal( SDL_Color *palette ); // load the palette
    void parseBin(TileList *bindata);

    SDL_Surface *loadTile( INIFile *templini,
                           Uint16 templ, Uint8 tile, Uint32 &tiletype ); // load a specified tile



    /* Some variables */

    Uint16 width;       // width of map
    Uint16 height;      // height of map
    Uint16 xpos;           // current xpos
    Uint16 ypos;           // current ypos
    Uint16 x;         // x coordinat for the first tile
    Uint16 y;         // y coordinat for the first tile

    char *theater;    // the theater of the map (eg. winter)
    char *brief;      // the briefing movie
    char *action;     // the actionmovie
    char *player;     // the players side (goodguy, badguy etc.)
    char *theme;
    char *winmov;
    char *losemov;

    char mapname[9];    // the name of the map

    /* A array of tiles and a vector constining the images for the tiles */

    Uint16 *tilematrix; // a matrix of the tiles
    std::vector<SDL_Surface *>tileimages;

    Uint16 *overlaymatrix;
    Uint8 *terraintypes;

    std::vector<SHPImage *> imagepool;
    std::map<Uint32, TerrainEntry> terrains;
    std::map<Uint32, Uint16> overlays;

    Uint8 maptype; // 0: TD 1: RA

    SDL_Surface *minimap;

    ActionEventQueue *evQueue;

    ScrollVector scrollvec;
    bool toscroll;

    UnitAndStructurePool *uspool;
    PlayerPool ppool;
};

#define SCROLL_NONE 0
#define SCROLL_UP 0x1
#define SCROLL_RIGHT 0x2
#define SCROLL_DOWN 0x4
#define SCROLL_LEFT 0x8
#define SCROLL_UPRIGHT (SCROLL_UP | SCROLL_RIGHT)
#define SCROLL_DOWNRIGHT (SCROLL_DOWN | SCROLL_RIGHT)
#define SCROLL_DOWNLEFT (SCROLL_DOWN | SCROLL_LEFT)
#define SCROLL_UPLEFT (SCROLL_UP | SCROLL_LEFT)
#define SCROLL_DELTA 1
#define MAXSCROLL (SCROLL_DELTA * 1)
#define MAXTSCROLL 7

#define TTYPE_LAND 0
#define TTYPE_WATER 1
#define TTYPE_ROAD 2
#define TTYPE_ROCK 3
#define TTYPE_TREE 4
#define TTYPE_WATER_BLOCKED 5
#define TTYPE_OTHER_NONPASS 7


#endif

