#ifndef __PLAYERPOOL_H
#define __PLAYERPOOL_H

#include "SDL.h"
#include <vector>

class INIFile;
class PlayerPool;
class CnCMap;
class AI;
class Unit;
class Structure;

typedef enum{PS_GOOD, PS_BAD, PS_NEUTRAL, PS_SPECIAL, PS_UNDEFINED} PSIDE;

class Player {
public:
    Player(char *pname, INIFile *mapini, PlayerPool *ppool);
    ~Player();

    char *getName() {
        return playername;
    }
    PSIDE getSide() {
        return playerside;
    }
    PSIDE getOtherSide() {
        if (playerside == PS_GOOD)
            return PS_BAD;
        if (playerside == PS_BAD)
            return PS_GOOD;
        return PS_UNDEFINED;
    }
    void increaseMoney(Uint32 change) {
        money += change;
    }
    bool decreaseMoney(Uint32 change);
    Uint32 getMoney() {
        return money;
    }

    void builtUnit(Unit* un);
    void lostUnit(Unit* un, bool wasDeployed);
    Uint16 getNumUnits() {
        return unitpool.size();
    }
    void builtStruct(Structure* str);
    void lostStruct(Structure* str);

    Uint16 getNumStruct() {
        return structurepool.size();
    }

    Uint8 getPaletteNum() {
        return palnum;
    }
    bool isAllied();
    Uint8 getNumAllies() {
        return numallies;
    }

    Uint32 getPower() {
        return pwr;
    }
    Uint32 getPowerUsed() {
        return pwrused;
    }
    AI *getAI() {
        return myAI;
    }
    void createAI(CnCMap *theMap);
    void updateOwner(Uint8 newnum);
    std::vector<Unit*> getUnits() {
        return unitpool;
    }
    std::vector<Structure*> getStructures() {
        return structurepool;
    }
private:
    char *playername;
    PSIDE playerside;
    //    Uint8 playernum;
    Uint8 palnum;

    Uint32 money;
    Uint32 pwr;
    Uint32 pwrused;

    std::vector<Unit*> unitpool;
    std::vector<Structure*> structurepool;

    Uint8 numallies;
    PlayerPool *ppool;
    AI *myAI;
};

class PlayerPool {
public:
    PlayerPool();
    ~PlayerPool();
    Uint8 getNumPlayers() {
        return playerpool.size();
    }
    Uint8 getHPlayerNum() {
        return humanPlayer;
    }
    Player *getHPlayer() {
        return playerpool[humanPlayer];
    }
    void setHPlayer(char *pname);
    Player *getPlayer(Uint8 player) {
        return playerpool[player];
    }
    Uint8 getEnemyNum(Player* pl) {
        // this assumes that the enemy exists
        return getOwnersForSide(pl->getOtherSide())[0];
    }
    Uint8 getPlayerNum(char *pname, INIFile *mapini);
    Uint8 getPlayerNum(Player *thePlayer);
    Uint8 getPaletteNum(Uint8 player) {
        return playerpool[player]->getPaletteNum();
    }
    std::vector<Uint8> getOwnersForSide(PSIDE thePlayerSide);
    void playerDefeated(Player *pl);
    bool hasWon() {
        return won;
    }
    bool hasLost() {
        return lost;
    }
private:
    std::vector<Player *> playerpool;
    Uint8 humanPlayer;
    bool won, lost;
};

// for the powerinfo_t struct definition
#include "structure.h"

#endif
