/*****************************************************************************
 * input.cpp - Class to handle all input from the user 
 *
 ****************************************************************************/

#include "input.h"
#include "unit.h"
#include "structure.h"
#include "soundengine.h"
#include "memtrack.h"
#include "ai.h"
#include "message.h"

#define MIN(x, y) ((x<y)?x:y)

bool Input::drawing;
SDL_Rect Input::markrect;
bool Input::minimapEnabled;


/** Constructor, sets up the input handeler.
 * @param the sidebar.
 * @param the map.
 * @param the width of the screen.
 * @param the height of the screen.
 */
Input::Input(Sidebar *sbar, CnCMap *map, Uint16 screenwidth,
             Uint16 screenheight, SDL_Rect *maparea, Cursor *cur, SoundEngine *sndengine) {
    this->map = map;
    this->sidebar = sbar;
    width = screenwidth;
    height = screenheight;
    done = 0;
    donecount = 0;
    this->maparea = maparea;
    tabwidth = sbar->getTabLocation()->w;
    tilewidth = map->getMapTile(0)->w;
    this->sndengine = sndengine;
    drawing = false;
    selected = new Selection();
    kbdmod = 0;
    lmousedown = MDOWN_NONE;
    rmousedown = MDOWN_NONE;

    cursor = cur;
    deploying_super = false;
    placing_struct = false;
    minimapEnabled = false;
    // pander
    cerr << "      F9  >>  Activates AI for human player  <<" << endl;
    cerr << "SHIFT-F9  >>  Activates AI for all non-human players  <<" << endl;
    cerr << "      F12 >>  print all Players, Units and Structures <<" << endl;
}

Input::~Input() {
    delete selected;
}

/** Method to handle the input events. */
void Input::handle() {
    SDL_Event event;
    int mx, my;
    Uint8 sdir;
    Uint8* keystate;

    while ( SDL_PollEvent(&event) ) {
        switch (event.type) {
        case SDL_MOUSEBUTTONDOWN: /* start drawing the selection square if in map */
            //	 SDL_GetMouseState(&mx, &my);
            mx = event.button.x;
            my = event.button.y;
            if( mx >= maparea->x && mx < maparea->x + maparea->w &&
                    my >= maparea->y && my < maparea->y + maparea->h ) {
                if( event.button.button == SDL_BUTTON_LEFT ) {
                    lmousedown = MDOWN_MAP;
                    markrect.x = mx;
                    markrect.y = my;
                } else if( event.button.button == SDL_BUTTON_RIGHT ) {
                    rmousedown = MDOWN_MAP;
                    drawing = false;
                    lmousedown = MDOWN_NONE;
                }
            }
            if (my < maparea->y + maparea->h && mx > width-tabwidth && sidebar->getVisible()) {
                clickSidebar(mx,my);
            }
            break;
        case SDL_MOUSEBUTTONUP:
            //	 SDL_GetMouseState(&mx, &my);
            mx = event.button.x;
            my = event.button.y;
            sidebar->resetButton();
            if( drawing ) {
                selectRegion();
            } else if( my >= maparea->y ) {
                if( my < maparea->y + maparea->h ) { /* map || sidebar */
                    if( mx >= maparea->x && mx < maparea->x+maparea->w ) {
                        if( event.button.button == SDL_BUTTON_LEFT && lmousedown == MDOWN_MAP) {
                            clickMap(mx, my);
                        } else if( event.button.button == SDL_BUTTON_RIGHT && rmousedown == MDOWN_MAP)
                            selected->clearSelection();
                    }
                    /* should also check if minimap is clicked,
                    will add when it's drawn */
                }
            } else { /* clicked a tab */
                if( mx < tabwidth )
                    printf("Options menu\n");
                if( mx > width-tabwidth )
                    sidebar->toggleVisible();
            }

            /* We can't be drawing now */
            lmousedown = MDOWN_NONE;
            rmousedown = MDOWN_NONE;
            drawing = false;
            break;
            /* A key has been pressed or released */
        case SDL_KEYDOWN:
            /* If it wasn't a press, ignore this event */
            if( event.key.state != SDL_PRESSED )
                break;

            /* Otherwise check which key was pressed */
            switch( event.key.keysym.sym ) {
            case SDLK_ESCAPE:
                /* It was esc so quit */
                done = 1;
                break;
            case SDLK_RSHIFT:
            case SDLK_LSHIFT:
                kbdmod |= KBDMOD_SHIFT;
                break;
            case SDLK_RCTRL:
            case SDLK_LCTRL:
                kbdmod |= KBDMOD_CTRL;
                break;
            case SDLK_RALT:
            case SDLK_LALT:
                kbdmod |= KBDMOD_ALT;
                break;
            case SDLK_TAB:
                sidebar->toggleVisible();
                break;
            case SDLK_F8:
                map->getUASP()->showMoves();
                break;
            case SDLK_F9: // pander for testing AI
                if ((kbdmod & KBDMOD_SHIFT) != KBDMOD_SHIFT) {
                    cerr << ">>  AI activated for human player  <<" << endl;
                    Player *aPlayer = map->getPlayerPool()->getHPlayer();
                    if (!aPlayer->getAI()) {
                        aPlayer->createAI(map);
                    }
                    aPlayer->getAI()->attackOtherSide();
                } else {
                    cerr << ">>  AI activated for all non-human players  <<" << endl;
                    for(Uint8 i = 0;
                            i < map->getPlayerPool()->getNumPlayers();
                            i++) {
                        Player *aPlayer = map->getPlayerPool()->getPlayer(i);
                        if (aPlayer->getSide() == map->getPlayerPool()->getHPlayer()->getOtherSide()) {
                            if (!aPlayer->getAI()) {
                                aPlayer->createAI(map);
                            }
                            aPlayer->getAI()->attackOtherSide();
                        }
                    }
                }
                break;
            case SDLK_F10: // pander for testing AI
                break;
            case SDLK_F11: // pander for testing AI
                break;
            case SDLK_F12: // pander for testing AI
                cerr << ">> print all Players, Units and Structures <<" << endl;
                map->printPlayersUnitsStructures();
                break;
            case SDLK_r:
                sidebar->startRadarAnim(0,&minimapEnabled);
                break;
            case SDLK_f:
                minimapEnabled = false;
                sidebar->startRadarAnim(1,&minimapEnabled);
                break;
            default:
                break;
            }
            break;

        case SDL_KEYUP:
            if( event.key.state != SDL_RELEASED )
                break;

            switch( event.key.keysym.sym ) {
            case SDLK_RSHIFT:
            case SDLK_LSHIFT:
                kbdmod &= ~KBDMOD_SHIFT;
                break;
            case SDLK_RCTRL:
            case SDLK_LCTRL:
                kbdmod &= ~KBDMOD_CTRL;
                break;
            case SDLK_RALT:
            case SDLK_LALT:
                kbdmod &= ~KBDMOD_ALT;
                break;
            default:
                break;
            }
            break;

            /* We got the quit event so quit */
        case SDL_QUIT:
            done = 1;
            break;
            /* By default we just ignore the event */
        default:
            break;
        }
    }
    sdir = 0;
    keystate = SDL_GetKeyState(NULL);
    if (keystate[SDLK_LEFT])
        sdir |= SCROLL_LEFT;
    else if (keystate[SDLK_RIGHT])
        sdir |= SCROLL_RIGHT;
    if (keystate[SDLK_UP])
        sdir |= SCROLL_UP;
    else if (keystate[SDLK_DOWN])
        sdir |= SCROLL_DOWN;
    if (sdir != SCROLL_NONE)
        map->scroll(sdir);

    if (keystate[SDLK_PAGEDOWN])
        sidebar->scrollSidebar(false);
    else if (keystate[SDLK_PAGEUP])
        sidebar->scrollSidebar(true);

    if( map->getUASP()->hasDeleted() ) {
        selected->checkSelection();
    }

    updateMousePos();

    if( map->getPlayerPool()->hasWon() || map->getPlayerPool()->hasLost() ) {
        ++donecount;
    }
    if (donecount == 1) {
        if (map->getPlayerPool()->hasWon()) {
            sndengine->queueSound("accom1.aud");
            MessagePool::currentInstance->postMessage("MISSION ACCOMPLISHED");
        } else {
            sndengine->queueSound("fail1.aud");
            MessagePool::currentInstance->postMessage("MISSION FAILED");
        }
    }
    if (donecount > 100) {
        done = 1;
    }
}

void Input::updateMousePos() {
    Uint8 scroll;
    Uint16 cursornum;
    Uint16 oldxscroll, oldyscroll;
    /* Check the position of the mousepointer and scroll if we are
     * less than 10 pixels from an edge. */
    int mx, my;
    SDL_GetMouseState(&mx, &my);


    if( drawing ) { /* set cursor to the default one when drawing */
        if( mx < maparea->x )
            markrect.w = maparea->x;
        else if( mx >= maparea->x+maparea->w )
            markrect.w = maparea->x+maparea->w-1;
        else
            markrect.w = mx;
        if( my < maparea->y )
            markrect.h = my = maparea->y;
        else if( my >= maparea->y+maparea->h )
            markrect.h = maparea->y+maparea->h-1;
        else
            markrect.h = my;

        cursor->setCursor(CUR_STANDARD, CUR_NOANIM);
    }
    /* Check if we should start drawing */
    else if( lmousedown == MDOWN_MAP && !placing_struct && !deploying_super) {
        if( my >= maparea->y && my < maparea->y+maparea->h &&
                mx >= maparea->x && mx < maparea->x+maparea->w )
            if(abs(markrect.x - mx) > 2 || abs(markrect.y - my) > 2) {
                drawing = true;
                markrect.w = mx;
                markrect.h = my;
            }
    } else { /* set the correct cursor at the end of this else */
        scroll = 0;
        if( mx > width - 10 )
            scroll |= SCROLL_RIGHT;
        else if( mx < 10 )
            scroll |= SCROLL_LEFT;
        if( my > height - 10 )
            scroll |= SCROLL_DOWN;
        else if( my < 10 )
            scroll |= SCROLL_UP;

        if( scroll != 0 ) {
            switch(scroll) {
            case SCROLL_UP:
                cursornum = CUR_SCROLLUP;
                break;
            case SCROLL_UPLEFT:
                cursornum = CUR_SCROLLUL;
                break;
            case SCROLL_LEFT:
                cursornum = CUR_SCROLLLEFT;
                break;
            case SCROLL_DOWNLEFT:
                cursornum = CUR_SCROLLDL;
                break;
            case SCROLL_DOWN:
                cursornum = CUR_SCROLLDOWN;
                break;
            case SCROLL_DOWNRIGHT:
                cursornum = CUR_SCROLLDR;
                break;
            case SCROLL_RIGHT:
                cursornum = CUR_SCROLLRIGHT;
                break;
            case SCROLL_UPRIGHT:
                cursornum = CUR_SCROLLUR;
                break;
            default:
                cursornum = CUR_STANDARD;
            }
            oldxscroll = map->getXScroll();
            oldyscroll = map->getYScroll();

            map->scroll(scroll);

            if( oldxscroll == map->getXScroll() && oldyscroll == map->getYScroll() )
                cursornum += CUR_NOSCROLL_OFFSET;
            if (!deploying_super) {
                cursor->setCursor(cursornum, CUR_NOANIM);
            }
        } else if( placing_struct ) {
            checkPlace(mx, my);
            return;
        } else if (deploying_super) {
            cursor->setXY(mx, my);
            return;
        } else {
            setCursorNum(mx, my);
        }
    }
    cursor->setXY(mx, my);
}

void Input::clickMap(int mx, int my) {
    Unit *curunit;
    Unit* tmpunit = NULL;
    Structure *curstructure;
    Uint16 pos;
    Uint8 subpos;
    bool sndplayed = false;
    bool enemy;

    clickedTile(mx, my, pos, subpos);
    if( pos == 0xffff )
        return;

    if( placing_struct ) {
        pos = checkPlace(mx, my);
        if( pos != 0xffff ) {
            placing_struct = false;
            if( temporary_place_unit ) {
                if (kbdmod == KBDMOD_CTRL) {
                    map->getUASP()->createUnit(placename, pos, 0, map->getPlayerPool()->getEnemyNum(
                                                   map->getPlayerPool()->getHPlayer()));
                } else {
                    map->getUASP()->createUnit(placename, pos, 0, map->getPlayerPool()->getHPlayerNum());
                }
            } else {
                map->getUASP()->createStructure(placename, pos, map->getPlayerPool()->getHPlayerNum(),true);
                sndengine->queueSound("hvydoor1.aud");
                if(!( (strcmp("sbag",placename) == 0) ||
                        (strcmp("brik",placename) == 0) ||
                        (strcmp("cycl",placename) == 0) ||
                        (strcmp("wood",placename) == 0) ||
                        (strcmp("barb",placename) == 0))) {
                    sndengine->queueSound("constru2.aud");
                }
            }
        }
        return;
    } else if (deploying_super) {
        deploying_super = false;
        if (strcmp("ion",placename) == 0) {
            sndengine->queueSound("ion1.aud");
        }
        return;
    }

    curunit = map->getUASP()->getUnitAt(pos, subpos);
    curstructure = map->getUASP()->getStructureAt(pos,selected->getWall());
    if( curunit != NULL ) {
        enemy = curunit->getOwner() != map->getPlayerPool()->getHPlayerNum();
        if( enemy ) {
            if( selected->canAttack() ) {
                selected->attackUnit(curunit);
                tmpunit = selected->getRandomUnit();
                if (tmpunit != NULL) {
                    sndengine->queueSound(((UnitType *)tmpunit->getType())->getRandTalk(2));
                }
            } else if( selected->empty() || selected->isEnemy() ) {
                selected->clearSelection();
                selected->addUnit(curunit, true);
            }
            return;
        } else if( kbdmod == KBDMOD_CTRL ) {
            if( selected->canAttack() ) {
                selected->attackUnit(curunit);
                tmpunit = selected->getRandomUnit();
                if (tmpunit != NULL) {
                    sndengine->queueSound(((UnitType *)tmpunit->getType())->getRandTalk(2));
                }
            }
        } else if( kbdmod == KBDMOD_SHIFT ) {
            if( selected->isEnemy() ) {
                selected->clearSelection();
            }
            if( curunit->isSelected() )
                selected->removeUnit(curunit);
            else {
                selected->addUnit(curunit, false);
                if (!sndplayed) {
                    sndengine->queueSound(((UnitType *)curunit->getType())->getRandTalk(0));
                    sndplayed = true;
                }
            }
        } else if( !curunit->isSelected() ) {
            selected->clearSelection();
            selected->addUnit(curunit, false);
            if (!sndplayed) {
                sndengine->queueSound(((UnitType *)curunit->getType())->getRandTalk(0));
                sndplayed = true;
            }
        } else if ((!selected->isEnemy())&&(curunit->canDeploy())) {
            selected->removeUnit(curunit);
            curunit->deploy();
        }
    } else if( curstructure != NULL ) {
        enemy = curstructure->getOwner() != map->getPlayerPool()->getHPlayerNum();
        if( enemy ) {
            if( selected->canAttack() ) {
                selected->attackStructure(curstructure);
                tmpunit = selected->getRandomUnit();
                if (tmpunit != NULL) {
                    sndengine->queueSound(((UnitType *)tmpunit->getType())->getRandTalk(3));
                }
            } else if( selected->empty() || selected->isEnemy() && !curstructure->isWall()) {
                selected->clearSelection();
                selected->addStructure(curstructure, true);
            }
            return;
        }
        if( kbdmod == KBDMOD_CTRL ) {
            if( selected->canAttack() ) {
                selected->attackStructure(curstructure);
                tmpunit = selected->getRandomUnit();
                if (tmpunit != NULL) {
                    sndengine->queueSound(((UnitType *)tmpunit->getType())->getRandTalk(3));
                }
            }
        } else if( kbdmod == KBDMOD_SHIFT ) {
            if( selected->isEnemy() ) {
                selected->clearSelection();
            }
            if( curstructure->isSelected() )
                selected->removeStructure(curstructure);
            else {
                if (!curstructure->isWall())
                    selected->addStructure(curstructure, false);
            }
        } else if( kbdmod == KBDMOD_ALT ) {
            curstructure->runSecAnim(5);
        } else if( !curstructure->isSelected() ) {
            selected->clearSelection();
            if (!curstructure->isWall())
                selected->addStructure(curstructure, false);
        }
    } else {
        if( selected->canMove() && map->getCost(pos) < 0xfff0) {
            if (!sndplayed) {
                sndengine->queueSound(((UnitType *)selected->getRandomUnit()->getType())->getRandTalk(1));
                sndplayed = true;
            }
            selected->moveUnits(pos);
        }
    }
}

void Input::clickedTile(int mx, int my, Uint16 &pos, Uint8 &subpos) {
    Uint16 xrest, yrest, tx, ty;
    mx -= maparea->x;
    my -= maparea->y;
    tx = mx/tilewidth;
    ty = my/tilewidth;
    if( tx >= map->getWidth() || ty >= map->getHeight() )
        pos = 0xffff;
    else {
        pos = (my/tilewidth)*map->getWidth()+mx/tilewidth;
        pos += map->getScrollPos();
    }

    xrest = mx%tilewidth;
    yrest = my%tilewidth;
    subpos = 0;
    if( xrest >= 8 && xrest < 16 && yrest >= 8 && yrest < 16 )
        return;

    subpos = 1;
    if( yrest >= 12 )
        subpos+=2;
    if( xrest >= 12 )
        subpos++;
    /* assumes the subposes are:
     * 1 2
     *  0
     * 3 4
     */
}

void Input::setCursorNum(int mx, int my) {
    Uint16 pos;
    Uint8 subpos;
    Unit *curunit;
    Structure *curstruct;
    bool enemy;

    /* clicked the map */
    if( my >= maparea->y && my < maparea->y+maparea->h &&
            mx >= maparea->x && mx < maparea->x+maparea->w ) {
        clickedTile(mx, my, pos, subpos);
        if( pos != 0xffff ) {
            curunit = map->getUASP()->getUnitAt(pos, subpos);
            curstruct = map->getUASP()->getStructureAt(pos,selected->getWall());
            if( curunit != NULL ) {
                enemy = curunit->getOwner() != map->getPlayerPool()->getHPlayerNum();
                if( !curunit->isSelected() ) {
                    if( selected->canAttack() && (enemy || (kbdmod == KBDMOD_CTRL))) {
                        cursor->setCursor("attack");
                        return;
                    } else if( !enemy || selected->empty() || selected->isEnemy() ) {
                        cursor->setCursor("select");
                        return;
                    }
                } else {
                    if ((!enemy)&&((UnitType*)curunit->getType())->canDeploy()) {
                        if (curunit->canDeploy())
                            cursor->setCursor("deploy");
                        else
                            cursor->setCursor(CUR_NOMOVE, CUR_NOANIM);
                        return;
                    }
                }
            } else if( curstruct != NULL ) {
                enemy = curstruct->getOwner() != map->getPlayerPool()->getHPlayerNum();
                if( !curstruct->isSelected() ) {
                    if( selected->canAttack() && (enemy || (kbdmod == KBDMOD_CTRL))) {
                        cursor->setCursor("attack");
                        return;
                    } else if( !enemy || selected->empty() || selected->isEnemy()) {
                        if (curstruct->isWall())
                            cursor->setCursor(CUR_NOMOVE, CUR_NOANIM);
                        else
                            cursor->setCursor("select");
                        return;
                    }
                }
            } else if( selected->canMove() ) {
                if( map->getCost(pos) < 0xfff0 ) {
                    cursor->setCursor(CUR_MOVE, CUR_NOANIM);
                } else {
                    cursor->setCursor(CUR_NOMOVE, CUR_NOANIM);
                }
                return;
            }
        }
    }


    cursor->setCursor(CUR_STANDARD, CUR_NOANIM);
}

void Input::selectRegion() {
    Uint16 startx, starty, stopx, stopy;
    Uint16 scannerx, scannery, curpos, i;
    Unit *un;
    bool playedsnd = false;

    selected->clearSelection();

    startx = (MIN(markrect.x, markrect.w)-maparea->x)/tilewidth+map->getXScroll();
    starty = (MIN(markrect.y, markrect.h)-maparea->y)/tilewidth+map->getYScroll();
    stopx = (abs(markrect.x - markrect.w)+1)/tilewidth+startx;
    stopy = (abs(markrect.y - markrect.h)+1)/tilewidth+starty;
    if( stopx >= map->getWidth() )
        stopx = map->getWidth() - 1;
    if( stopy >= map->getHeight() )
        stopy = map->getHeight() - 1;

    for( scannery = starty; scannery <= stopy; scannery++ ) {
        for( scannerx = startx; scannerx <= stopx; scannerx++ ) {
            curpos = scannery*map->getWidth()+scannerx;
            for( i = 0; i < 5; i++ ) {
                un = map->getUASP()->getUnitAt(curpos, i);
                if( un != NULL ) {
                    if( un->getOwner() != map->getPlayerPool()->getHPlayerNum() ) {
                        continue;
                    }
                    selected->addUnit(un, false);
                    if (!playedsnd) {
                        sndengine->queueSound(((UnitType *)un->getType())->getRandTalk(0));
                        playedsnd = true;
                    }
                    if( !((UnitType *)un->getType())->isInfantry() )
                        break;
                }
            }
        }
    }

}

void Input::clickSidebar(int mx, int my) {
    Uint8 createmode, butclick;

    mx -= (width-tabwidth);
    my -= maparea->y;
    butclick = sidebar->getButton(mx,my);
    if (butclick != 255) {
        /** @todo find a more elegant way to do this, as scrolling will blank
         *  current place.
         */
        strncpy(placename,"xxxx",4);
        sidebar->clickButton(butclick,placename,&createmode);

        if( (strncasecmp("xxxx",placename,4) == 0) ) {
            placing_struct = false;
            deploying_super = false;
            return;
        }
        if( (strncasecmp("atom",placename,4) == 0)
                || (strncasecmp("ion",placename,3) == 0) ||(strncasecmp("bomb",placename,4) == 0)) {
            placing_struct = false;
            deploying_super = true;
            if ( (strncasecmp("atom",placename,4) == 0) || (strncasecmp("ion",placename,3) == 0)) {
                cursor->setCursor(placename);
            } else {
                cursor->setCursor("superdeploy");
            }
            sndengine->queueSound("select1.aud");
            return;// was scrolling or clicked on super weap
        }
        /* ugly++, place units as structures */
        if( createmode ) {
            placing_struct = true;
            deploying_super = false;
            placeposvalid = true;
            temporary_place_unit = true;
            placetype = map->getUASP()->getStructureTypeByName("sbag");
            return;
        }
        temporary_place_unit = false;
        placing_struct = true;
        deploying_super = false;
        placeposvalid = true;
        placetype = map->getUASP()->getStructureTypeByName(placename);
    }
}

Uint16 Input::checkPlace(int mx, int my) {
    Uint8 placexpos, placeypos;
    Uint8 *placemat;
    Uint16 pos;
    Uint8 subpos;

    if( my < maparea->y || my >= maparea->y+maparea->h ||
            mx < maparea->x || mx >= maparea->x+maparea->w ) {
        cursor->setCursor(CUR_STANDARD, CUR_NOANIM);
        cursor->setXY(mx, my);
        return 0xffff;
    }

    clickedTile(mx, my, pos, subpos);
    if( pos == 0xffff ) {
        cursor->setCursor(CUR_STANDARD, CUR_NOANIM);
        cursor->setXY(mx, my);
        placeposvalid = false;
        return pos;
    }

    if( pos%map->getWidth() + placetype->getXsize() - map->getXScroll() > MIN(maparea->w/tilewidth, map->getWidth())) {
        pos -= (pos%map->getWidth() + placetype->getXsize() - map->getXScroll() - MIN(maparea->w/tilewidth, map->getWidth()));
    }
    if( pos/map->getWidth() + placetype->getYsize() - map->getYScroll() > MIN(maparea->h/tilewidth, map->getHeight())) {
        pos -= ((pos/map->getWidth() + placetype->getYsize() - map->getYScroll() - MIN(maparea->h/tilewidth, map->getHeight())))*map->getWidth();
    }

    /* check if pos is valid and set cursor */
    placeposvalid = true;
    //	cursor->setCursor(CUR_STANDARD, CUR_NOANIM);
    placemat = new Uint8[placetype->getXsize()*placetype->getYsize()];
    for( placeypos = 0; placeypos < placetype->getYsize(); placeypos++ ) {
        for( placexpos = 0; placexpos < placetype->getXsize(); placexpos++ ) {
            if( placetype->isBlocked(placeypos*placetype->getXsize()+placexpos) ) {
                placemat[placeypos*placetype->getXsize()+placexpos] = 1;
                if( map->getCost(pos+placeypos*map->getWidth()+placexpos)>1 ) {
                    placeposvalid = false;
                    placemat[placeypos*placetype->getXsize()+placexpos] = 2;
                }
            } else {
                placemat[placeypos*placetype->getXsize()+placexpos] = 0;
            }
        }
    }
    cursor->setPlaceCursor(placetype->getXsize(), placetype->getYsize(), placemat);

    delete[] placemat;
    drawing = false;

    mx = (pos%map->getWidth() - map->getXScroll())*tilewidth + maparea->x;
    my = (pos/map->getWidth() - map->getYScroll())*tilewidth + maparea->y;
    cursor->setXY(mx, my);
    if( !placeposvalid ) {
        pos = 0xffff;
    }
    return pos;
}
