/*****************************************************************************
 * sidebar.cpp - sidebar class
 *
 ****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include "SDL.h"

#include "sidebar.h"
#include "shpimage.h"
#include "playerpool.h"

#include "memtrack.h"

/** Constructor, creates the sidebar.
 * @param the mixfiles.
 */
Sidebar::Sidebar( ActionEventQueue* evQueue, Player *pl,
                  Uint16 height, char *theater) {

    SHPImage *tmpshp;
    SDL_Surface *shadow;

    scaleq = -1;

    this->theater = theater;
    player = pl;

    try {
        tmpshp = new SHPImage("htabs.shp", scaleq);
    } catch(ImageNotFound *e) {
        delete e;
        try {
            tmpshp = new SHPImage("tabs.shp", scaleq);
        } catch(ImageNotFound *e) {
            delete e;
            fprintf(stderr, "Unable to load the tab-image! (Mayby you run c&c gold but have forgoten updatec.mix?)\n");
            exit(1);
        }
    }


    tmpshp->getImage(0, &tab, &shadow, 0);
    SDL_FreeSurface(shadow);
    delete tmpshp;

    tablocation.x = 0;
    tablocation.y = 0;
    tablocation.w = tab->w;
    tablocation.h = tab->h;

    if (player->getSide() == PS_BAD) {
        strncpy(radarname,"hradar.nod",13);
    } else if (player->getSide() == PS_GOOD) {
        strncpy(radarname, "hradar.gdi",13);
    } else {
        strncpy(radarname, "hradar.mp",13);
    }

    try {
        tmpshp = new SHPImage(radarname, scaleq);
    } catch(ImageNotFound *e) {
        delete e;
        try {
            if (player->getSide() == PS_BAD) {
                strncpy(radarname,"radar.nod",13);
            } else if (player->getSide() == PS_GOOD) {
                strncpy(radarname, "radar.gdi",13);
            } else {
                strncpy(radarname, "radar.mp",12);
            }
            tmpshp = new SHPImage(radarname, scaleq);
        } catch(ImageNotFound *e) {
            delete e;
            fprintf(stderr, "Unable to load the radar-image! (Mayby you run c&c gold but have forgoten updatec.mix?)\n");
            exit(1);
        }
    }

    tmpshp->getImage(0, &radarlogo, &shadow, 0);
    SDL_FreeSurface(shadow);
    delete tmpshp;

    visible = 1;
    changed = 1;

    unitoff = 0;
    structoff = 0;

    buttondown = 0;
    bd = false;

    sbar = NULL;
    sbarlocation.x = 0;

    this->evQueue = evQueue;

    this->radaranim = NULL;
    radaranimating = false;

    setupbuttons(height);

    gamefnt = new Font("scorefnt.fnt");
}

/** Destructor, free all memory used by the sidebar.*/
Sidebar::~Sidebar() {
    Uint32 i;
    SDL_FreeSurface(tab);
    SDL_FreeSurface(radarlogo);
    if( sbar != NULL )
        SDL_FreeSurface(sbar);
    delete gamefnt;
    for( i = 0; i < buttons.size(); i++ ) {
        delete buttons[i];
    }
}

/** Check if the sidebar has changed and needs to be redrawn.
 * By changed we mean has become visible or hidden.
 * @returns 1 if the sidebar has changed, 0 otherwise.
 */
Uint8 Sidebar::getChanged() {
    if( changed > 0 ) {
        changed = 0;
        return 1;
    }
    return 0;
}

/** Toggle sidebar visible. */
void Sidebar::toggleVisible() {
    if( visible == 1 )
        visible = 0;
    else
        visible = 1;
    changed = 1;
}

SDL_Surface *Sidebar::getSidebarImage(SDL_Rect location) {
    SDL_Rect dest, src;
    SDL_Surface *temp, *shadow;
    SHPImage *textureshp;


    if( location.w == sbarlocation.w && location.h == sbarlocation.h )
        return sbar;

    if( sbar != NULL )
        SDL_FreeSurface(sbar);

    temp = SDL_CreateRGBSurface(SDL_SWSURFACE, location.w, location.h, 16, 0, 0, 0, 0);
    sbar = SDL_DisplayFormat(temp);
    SDL_FreeSurface(temp);
    location.x = 0;
    location.y = 0;

    try {
        textureshp = new SHPImage("btexture.shp", scaleq);
        textureshp->getImage(1, &temp, &shadow, 0);
        delete textureshp;
        SDL_FreeSurface(shadow);
        dest.x = 0;
        dest.w = location.w;
        dest.h = temp->h;
        src.x = 0;
        src.y = 0;
        src.h = temp->h;
        src.w = location.w;
        for( dest.y = 0; dest.y < location.h; dest.y += dest.h ) {
            SDL_BlitSurface( temp, &src, sbar, &dest );
            //      printf("textureblit\n");
        }
        SDL_FreeSurface(temp);
    } catch(ImageNotFound *e) {
        delete e;
        SDL_FillRect(sbar, &location, SDL_MapRGB(sbar->format, 0xa0, 0xa0, 0xa0));
    }
    sbarlocation = location;

    dest.x = 0;
    dest.y = 0;
    dest.w = radarlogo->w;
    dest.h = radarlogo->h;
    SDL_BlitSurface( radarlogo, NULL, sbar, &dest);

    for (Uint8 x=0;x<buttons.size();++x) {
        dest = buttons[x]->getRect();
        temp = buttons[x]->getSurface();
        SDL_BlitSurface( temp, NULL, sbar, &dest);
    }

    return sbar;
}

void Sidebar::addButton(Uint16 x, Uint16 y, char* fname, Uint8 f) {
    SidebarButton *tmp = new SidebarButton(x,y,fname,f,theater);
    buttons.push_back(tmp);
    changed = 1;
}

void Sidebar::setupbuttons(Uint16 height) {
    Uint16 scrollbase;
    Uint8 t;

    Uint32 striph,stripw;
    Uint32 startoffs = tab->h+radarlogo->h;

    SHPImage *strip;

    try {
        strip = new SHPImage("hstrip.shp", scaleq);
    } catch(ImageNotFound *e) {
        delete e;
        try {
            strip = new SHPImage("strip.shp", scaleq);
        } catch(ImageNotFound *e) {
            delete e;
            fprintf(stderr, "Unable to load strip images for sidebar, exiting\n");
            exit(1);
        }
    }

    striph = strip->getHeight();
    stripw = strip->getWidth();

    delete strip;

    if( striph > 100 )
        striph = striph>>2;

    buildbut = ((height-startoffs)/striph)-2;
    startoffs += striph;

    scrollbase = startoffs + striph*buildbut;
    addButton(10+stripw,scrollbase,"stripup.shp",B_SCROLL|B_UNIT|B_UP); // 0
    addButton(10,scrollbase,"stripup.shp",B_SCROLL|B_STRUCTURE|B_UP); // 1

    addButton(10+stripw+(stripw>>1),scrollbase,"stripdn.shp",B_SCROLL|B_UNIT|B_DOWN); // 2
    addButton(10+(stripw>>1),scrollbase,"stripdn.shp",B_SCROLL|B_STRUCTURE|B_DOWN); // 3

    // The order in which the addButton calls are made MUST be preserved
    // Two loops are made so that all unit buttons and all structure buttons
    // are grouped contiguously (4,5,6,7,...) compared to (4,6,8,10,...)

    for (t=0;t<buildbut;++t) {
        addButton(10+stripw,startoffs+striph*t,"strip.shp",B_BUILD|B_UNIT);
    }
    for (t=0;t<buildbut;++t) {
        addButton(10,startoffs+striph*t,"strip.shp",B_BUILD|B_STRUCTURE);
    }

    createdummylists();

    updateicons();
}

Uint8 Sidebar::getButton(Uint16 x,Uint16 y) {
    SDL_Rect tmp;

    for (Uint8 i=0;i<buttons.size();++i) {
        tmp = buttons[i]->getRect();
        if (x>=tmp.x && y>=tmp.y && x<(tmp.x+tmp.w) && y<(tmp.y+tmp.h)) {
            return i;
        }
    }
    return 255;
}

void Sidebar::clickButton(Uint8 index, char* unitname, Uint8* createmode) {
    Uint8 f = buttons[index]->getFunction();

    switch (f&0x3) {
    case 0:
        return;
        break;
    case 1: // build
        build(index - 3 - ((f&B_UNIT)?0:buildbut),(f&B_UNIT),unitname, createmode);
        break;
    case 2: // scroll
        downbutton(index);
        scrollbuildlist( (f&B_UP), (f&B_UNIT) );
        break;
    default:
        fprintf(stderr,"We got a %i coming through.  This should not happen\n",f&0x3);
        break;
    }
}

void Sidebar::build(Uint8 index, Uint8 type, char* unitname, Uint8* createmode) {
    Uint8* offptr;
    std::vector<char*>* vecptr;

    if (type) {
        offptr = &unitoff;
        vecptr = &uniticons;
        *createmode = 1;
    } else {
        offptr = &structoff;
        vecptr = &structicons;
        *createmode = 0;
    }

    //    printf("Building a %s, item %i on list\n",(type)?"unit":"structure",index);

    if ( (unsigned)(*vecptr).size() > ((unsigned)(*offptr)+index - 1) ) {
        strncpy(unitname,(*vecptr)[(*offptr+index-1)],13);
        unitname[strlen((*vecptr)[(*offptr+index-1)])-8] = 0x0;
        //   printf("It is a: %s\n",unitname);
    }
    else {
        //  printf("No assigned icon\n");
    }
}

void Sidebar::scrollbuildlist(Uint8 dir, Uint8 type) {
    Uint8* offptr;
    std::vector<char*>* vecptr;

    if (type) {
        offptr = &unitoff;
        vecptr = &uniticons;
    } else {
        offptr = &structoff;
        vecptr = &structicons;
    }

    if (dir) { //up
        if (*offptr>0) {
            --(*offptr);
        }
    } else { // down
        if ((unsigned)(*offptr+buildbut) < (unsigned)vecptr->size()) {
            ++(*offptr);
        }
    }

    updateicons();
}

void Sidebar::createdummylists(void) {
    uniticons.push_back("e1icon.shp");
    uniticons.push_back("e2icon.shp");
    uniticons.push_back("e3icon.shp");
    uniticons.push_back("e4icon.shp");
    uniticons.push_back("e5icon.shp");
    uniticons.push_back("e6icon.shp");
    uniticons.push_back("rmboicon.shp");
    uniticons.push_back("bikeicon.shp");
    uniticons.push_back("bggyicon.shp");
    uniticons.push_back("jeepicon.shp");
    uniticons.push_back("apcicon.shp");
    uniticons.push_back("ltnkicon.shp");
    uniticons.push_back("mtnkicon.shp");
    uniticons.push_back("ftnkicon.shp");
    uniticons.push_back("stnkicon.shp");
    uniticons.push_back("htnkicon.shp");
    uniticons.push_back("artyicon.shp");
    uniticons.push_back("harvicon.shp");
    uniticons.push_back("ionicon.shp");
    uniticons.push_back("atomicon.shp");
    uniticons.push_back("bombicon.shp");
    uniticons.push_back("a10icon.shp");
    uniticons.push_back("orcaicon.shp");
    uniticons.push_back("tranicon.shp");
    uniticons.push_back("mcvicon.shp");
    uniticons.push_back("mlrsicon.shp");
    uniticons.push_back("msamicon.shp");

    structicons.push_back("nukeicon.shp");
    structicons.push_back("procicon.shp");
    structicons.push_back("sbagicon.shp");
    structicons.push_back("brikicon.shp");
    structicons.push_back("cyclicon.shp");
    structicons.push_back("barbicon.shp");
    structicons.push_back("woodicon.shp");
    structicons.push_back("siloicon.shp");
    structicons.push_back("hqicon.shp");
    structicons.push_back("eyeicon.shp");
    structicons.push_back("gunicon.shp");
    structicons.push_back("atwricon.shp");
    structicons.push_back("gtwricon.shp");
    structicons.push_back("nuk2icon.shp");
    structicons.push_back("weapicon.shp");
    structicons.push_back("handicon.shp");
    structicons.push_back("pyleicon.shp");
    structicons.push_back("samicon.shp");
    structicons.push_back("fixicon.shp");
    structicons.push_back("afldicon.shp");
    structicons.push_back("hpadicon.shp");
    structicons.push_back("obliicon.shp");
    structicons.push_back("tmplicon.shp");
}

void Sidebar::updateicons(void) {
    Uint8 i;
    SDL_Rect temp;

    for (i=0;i<buildbut;++i) {
        if ((unsigned)(i+unitoff)>=(unsigned)uniticons.size())
            buttons[4+i]->ChangeImage("strip.shp");
        else {
            buttons[4+i]->ChangeImage(uniticons[i+unitoff]);
        }
    }
    for (i=0;i<buildbut;++i) {
        if ((unsigned)(i+structoff)>=(unsigned)structicons.size())
            buttons[buildbut+4+i]->ChangeImage("strip.shp");
        else
            buttons[buildbut+4+i]->ChangeImage(structicons[i+structoff]);
    }

    for (Uint8 x=0;x<buttons.size();++x) {
        temp = buttons[x]->getRect();
        if (sbar != NULL)
            SDL_FillRect(sbar,&temp,0x0);
        SDL_BlitSurface(buttons[x]->getSurface(), NULL, sbar, &temp);
    }

}

void Sidebar::downbutton(Uint8 index) {
    if (index>3)
        return; // not a scroll button

    buttondown = index;

    bd = true;
    if (index == 0 || index == 1)
        buttons[index]->ChangeImage("stripup.shp",1);
    else
        buttons[index]->ChangeImage("stripdn.shp",1);
    updateicons();

}

void Sidebar::resetButton(void) {
    if (!bd)
        return;
    bd = false;

    if (buttondown < 2)
        buttons[buttondown]->ChangeImage("stripup.shp",0);
    else
        buttons[buttondown]->ChangeImage("stripdn.shp",0);
    buttondown = 0;
    updateicons();
}

void Sidebar::startRadarAnim(Uint8 mode, bool* minienable) {
    if (radaranimating == false && radaranim == NULL) {
        radaranimating = true;
        radaranim = new RadarAnimEvent(2,mode,this,minienable);
    }
}

void Sidebar::scrollSidebar(bool scrollup) {
    scrollbuildlist(scrollup, 0);
    scrollbuildlist(scrollup, 1);
}

SidebarButton::SidebarButton(Sint16 x, Sint16 y, char* mixname,
                             Uint8 f, char* theater) {
    pic = NULL;
    picloc.x = x;
    picloc.y = y;
    function = f;
    this->theater = theater;
    ChangeImage(mixname);
}

bool SidebarButton::ChangeImage(char* fname) {
    return ChangeImage(fname,0);
}

bool SidebarButton::ChangeImage(char* fname, Uint8 number) {
    SHPImage* tmpshp;
    SDL_Surface* shadow;

    int scaleq = -1;
    char goldname[128];

    //if (pic != NULL)
    //    SDL_FreeSurface(pic);

    if( strlen(fname)>8 && !strcmp("icon.shp", fname+strlen(fname)-8) ) {
        strcpy(goldname, fname);
        goldname[strlen(fname)-6] = 'n';
        goldname[strlen(fname)-5] = 'h';
        goldname[strlen(fname)-3] = theater[0];
        goldname[strlen(fname)-2] = theater[1];
        goldname[strlen(fname)-1] = theater[2];
        try {
            tmpshp = new SHPImage(goldname, scaleq);
        } catch( ImageNotFound *e) {
            delete e;
            try {
                tmpshp = new SHPImage(fname, scaleq);
            } catch( ImageNotFound *e) {

                fprintf(stderr, "Icon not found \"%s\"\n", fname);
                return false;
            }
        }
    } else {
        sprintf(goldname, "h%s", fname);
        try {
            tmpshp = new SHPImage(goldname, scaleq);
        } catch( ImageNotFound *e) {
            delete e;
            try {
                tmpshp = new SHPImage(fname, scaleq);
            } catch( ImageNotFound *e) {

                fprintf(stderr,"Image not found: \"%s\"\n",fname);
                return false;
            }
        }
    }

    tmpshp->getImage(number,&pic,&shadow,0);
    SDL_FreeSurface(shadow);
    delete tmpshp;

    picloc.w = pic->w;
    picloc.h = pic->h;

    return true;
}

SidebarButton::~SidebarButton() {
    SDL_FreeSurface(pic);
}


RadarAnimEvent::RadarAnimEvent(Uint32 p, Uint8 mode, Sidebar* sidebar, bool* minienable) : ActionEvent(p) {
    this->mode = mode;
    this->sidebar = sidebar;
    this->minienable = minienable;
    switch (mode) {
    case 0:
        frame = 0;
        framend = 20;
        break;
    case 1:
        //*minienable = false;
        frame = 20;
        framend = 30;
        break;
    default:
        frame = 0;
    }
    sidebar->evQueue->scheduleEvent(this);
}

void RadarAnimEvent::run() {
    SDL_Rect dest;
    SDL_Surface *shadow;
    SHPImage* tmpshp;

    if (frame <= framend) {
        tmpshp = new SHPImage(sidebar->radarname, sidebar->scaleq);

        dest.x = 0;
        dest.y = 0;
        dest.w = sidebar->radarlogo->w;
        dest.h = sidebar->radarlogo->h;

        SDL_FillRect(sidebar->radarlogo, &dest, SDL_MapRGB(sidebar->sbar->format, 0x0a, 0x0a, 0x0a));
        SDL_BlitSurface( sidebar->radarlogo, NULL, sidebar->sbar, &dest);

        tmpshp->getImage(frame, &(sidebar->radarlogo), &shadow, 0);
        SDL_BlitSurface( sidebar->radarlogo, NULL, sidebar->sbar, &dest);

        SDL_FreeSurface(shadow);
        delete tmpshp;
        ++frame;
        sidebar->evQueue->scheduleEvent(this);
    } else {
        dest.x = 0;
        dest.y = 0;
        dest.w = sidebar->radarlogo->w;
        dest.h = sidebar->radarlogo->h;

        if (mode == 1) {
            tmpshp = new SHPImage(sidebar->radarname, sidebar->scaleq);
            tmpshp->getImage(0, &(sidebar->radarlogo), &shadow, 0);
            SDL_BlitSurface( sidebar->radarlogo, NULL, sidebar->sbar, &dest);
            SDL_FreeSurface(shadow);
            delete tmpshp;
            sidebar->radaranim = NULL;
        } else {
            //SDL_FillRect(sidebar->radarlogo, &dest, SDL_MapRGB(sidebar->sbar->format, 0x0a, 0x0a, 0x0a));
            //SDL_BlitSurface( sidebar->radarlogo, NULL, sidebar->sbar, &dest);
            *minienable = true;
            sidebar->radaranim = NULL;
        }
        sidebar->radaranimating = false;
        delete this;
        return;
    }
}
