#include "imagecache.h"

#include "memtrack.h"

using namespace std;


/** Constructor, creates the imagecache and prepares it to be used. */
ImageCache::ImageCache() {
    cache = NULL;
    prevcache = NULL;
}

/** Destructor, free all mem used by the imagecache. */
ImageCache::~ImageCache() {
    map<Uint32, ImageCacheEntry*>::iterator cachepos;
    if( cache != NULL ) {
        for( cachepos = (*cache).begin(); cachepos != (*cache).end();cachepos++) {
            if( cachepos->second != NULL ) {
                SDL_FreeSurface( cachepos->second->image );
                SDL_FreeSurface( cachepos->second->shadow );
                delete cachepos->second;
            }
        }
        delete cache;
    }
    if( prevcache != NULL ) {
        for( cachepos = (*prevcache).begin(); cachepos != (*prevcache).end();cachepos++) {
            if( cachepos->second != NULL ) {
                SDL_FreeSurface( cachepos->second->image );
                SDL_FreeSurface( cachepos->second->shadow );
                delete cachepos->second;
            }
        }
        delete prevcache;
    }
}

/** Sets the imagepool the imagecache will use.
 * @param the imagepool to use.
 */
void ImageCache::setImagePool( vector<SHPImage *> *imagepool ) {
    map<Uint32, ImageCacheEntry*>::iterator cachepos;
    if( cache != NULL ) {
        for( cachepos = (*cache).begin(); cachepos != (*cache).end();cachepos++) {
            if( cachepos->second != NULL ) {
                SDL_FreeSurface( cachepos->second->image );
                SDL_FreeSurface( cachepos->second->shadow );
                delete cachepos->second;
            }
        }
        delete cache;
        cache = NULL;
    }
    if( prevcache != NULL ) {
        for( cachepos = (*prevcache).begin(); cachepos != (*prevcache).end();cachepos++) {
            if( cachepos->second != NULL ) {
                SDL_FreeSurface( cachepos->second->image );
                SDL_FreeSurface( cachepos->second->shadow );
                delete cachepos->second;
            }
        }
        delete prevcache;
        prevcache = NULL;
    }
    this->imagepool = imagepool;
}

/** Gets an image from the cache, loads it to the cache from the imagepool
 * if the image is not in the cache.
 * @param the number of the image in the imagepool.
 * @returns A class containing the image and the shadow.
 */
ImageCacheEntry *ImageCache::getImage(Uint32 imgnum) {
    ImageCacheEntry *entry;
    map<Uint32, ImageCacheEntry*>::iterator cachepos;

    if( cache == NULL ) {
        cache = new map<Uint32, ImageCacheEntry*>;
    }

    cachepos = cache->find(imgnum);
    if( cachepos != cache->end() ) {
        /* found the entry */
        return cachepos->second;
    } else {
        /* didn't find the entry so check the old cache */
        if( prevcache != NULL ) {
            printf("get cached\n");
            cachepos = prevcache->find(imgnum);
            if( cachepos != prevcache->end() ) {
                /* the image was in the old cache */
                entry = cachepos->second;
                (*cache)[imgnum] = entry;
                (*prevcache)[imgnum] = NULL;
                return entry;
            } else {
                /* we have to load the image */
                entry = new ImageCacheEntry;
                /* palette is ((imgnum>>11)&0x1f) */
                (*imagepool)[imgnum>>16]->getImage(imgnum&0x7FF, &(entry->image), &(entry->shadow),
                                                   ((imgnum>>11)&0x1f) );
                (*cache)[imgnum] = entry;
                return entry;
            }
        } else {
            /* we have to load the image */
            entry = new ImageCacheEntry;
            (*imagepool)[imgnum>>16]->getImage(imgnum&0x7FF, &(entry->image), &(entry->shadow),
                                               ((imgnum>>11)&0x1f));
            (*cache)[imgnum] = entry;
            return entry;
        }
    }

}

/** Move the cache to oldcache and create a new cache to use. */
void ImageCache::newCache() {
    map<Uint32, ImageCacheEntry*>::iterator cachepos;
    if( prevcache != NULL ) {
        for( cachepos = (*prevcache).begin(); cachepos != (*prevcache).end();cachepos++) {
            if( cachepos->second != NULL ) {
                SDL_FreeSurface( cachepos->second->image );
                SDL_FreeSurface( cachepos->second->shadow );
                delete cachepos->second;
            }
        }
        delete prevcache;
    }
    prevcache = cache;
    cache = NULL;
}
