#include "SDL.h"
#include <stdlib.h>
#include <string.h>
#include "memtrack.h"
#include "imageproc.h"

/** Constructor, empty
 *
 */
ImageProc::ImageProc() {}



/** Destructor, empty */
ImageProc::~ImageProc() {}


/** Scale the incoming image
 * @param the image to process
 * @param quality. i.e. what method should be used:
 	0 (default) no interpolation, just spreaded
 	1 no interpolation, non-interleaved
 	2 linear interpolation
 */


SDL_Surface* ImageProc::scale(SDL_Surface *input, int quality) {
    int bytesPerPixel = input->format->BytesPerPixel;
    if(bytesPerPixel !=1) {
        fprintf(stderr,"ImageProc::scale() !Warning! decoding a non 8bpp surface: %i bpp", bytesPerPixel *8);
    }

    SDL_Surface *output = SDL_CreateRGBSurface (SDL_SWSURFACE, (input->w*2),(input->h * 2), bytesPerPixel *8, 0xff, 0xff, 0xff, 0);
    output->format->Rmask = input->format->Rmask;
    output->format->Gmask = input->format->Gmask;
    output->format->Bmask = input->format->Bmask;
    output->format->Amask = input->format->Amask;
    if( input->format->palette != NULL ) {
        SDL_SetColors(output, input->format->palette->colors, 0, input->format->palette->ncolors);
    }
    Uint8 *pixelsIn = (Uint8 *)input->pixels;
    Uint8 *pixelsOut = (Uint8 *)output->pixels;
    //put the pixels of the input image in the output image with a space of 1
    //pixel between every pixel (horizontal as well as vertical)
    int i= 0;
    int j = 0;
    int lineCounter = 0;
    int limit = (input->w * input->h) * bytesPerPixel;
    for(i = 0; i < limit ;i+=bytesPerPixel, j+=2, lineCounter++) {
        if(lineCounter == input->w *bytesPerPixel) {
            lineCounter= 0;
            memcpy(pixelsOut + j,pixelsOut + j -output->w, output->w);
            j+= output->w;
        }
        pixelsOut[j] = pixelsIn [i];
        pixelsOut[j+1] = pixelsIn [i];

    }
    return output;
}


/*SDL_Surface* ImageProc::scale(SDL_Surface *input, int quality){
	int bytesPerPixel = input->format->BytesPerPixel;
	if(bytesPerPixel !=1){
		fprintf(stderr,"ImageProc::scale() !Warning! decoding a non 8bpp surface: %i bpp", bytesPerPixel *8);
	}
 
	SDL_Surface *output = SDL_CreateRGBSurface (SDL_SWSURFACE, (input->w*2),(input->h * 2), bytesPerPixel *8, 0xff, 0xff, 0xff, 0);
	output->format->Rmask = input->format->Rmask;
	output->format->Gmask = input->format->Gmask;
	output->format->Bmask = input->format->Bmask;
	output->format->Amask = input->format->Amask;
    if( input->format->palette != NULL ){
        SDL_SetColors(output, input->format->palette->colors, 0, input->format->palette->ncolors);
    }
    Uint8 *pixelsIn = (Uint8 *)input->pixels;
	Uint8 *pixelsOut = (Uint8 *)output->pixels;
	//put the pixels of the input image in the output image with a space of 1
    //pixel between every pixel (horizontal as well as vertical)
    int i= 0;
    int j = 0;
    int lineCounter = 0;
    int jFactor = 2 * bytesPerPixel;
    int limit = (input->w * input->h) * bytesPerPixel;
    for(i = 0; i < limit ;i+=bytesPerPixel, j+= jFactor, lineCounter++){
       	if(lineCounter == input->w *bytesPerPixel){
       		lineCounter= 0;
       		j+= output->w * bytesPerPixel;
       	}
       	pixelsOut[j] = pixelsIn [i];
       	//pixelsOut[j+1] = pixelsIn [i+1];
       	////////added simple scale code//////
       	pixelsOut[j+1] = pixelsIn [i];
       	pixelsOut[j+output->w] = pixelsIn[i];
       	pixelsOut[j+output->w +1] = pixelsIn[i];
       	////////end//////////////////////////
       }
 
        if(quality==0){
        	//nothing, the image will be interlaced
        }else if( quality == 1){
        	//simpleScale(output);
        }else if ( quality == 2){
        	linInt(output);
        }
 
 
        return output;
}
*/
/*
void ImageProc::simpleScale(SDL_Surface *img){
	Uint8 *pixels = (Uint8 *) img->pixels;
	int bpp = img->format->BytesPerPixel;
	Uint32 iMinusOne = 0;
	//Uint32 iMinusTwo = 0;
	int lineCounter =0;
	for(Sint32 i = bpp; i < (img->w * img->h); i+= 2 * bpp, lineCounter+=2 ){
        	if(lineCounter == img->w *bpp){
        		lineCounter= 0;
        		i+= img->w * bpp;
        	}
        	iMinusOne = i -1;
        	//iMinusTwo = i-2;
        	pixels[i] = pixels[iMinusOne]; //horizontal copying of pixels
        	//pixels[i+1] = pixels[iMinusOne];
        	pixels[i + img->w*bpp] = pixels[iMinusOne];         //diagonal copying of pixels, direction: SE
        	//pixels[i + img->w*2 +1] = pixels[iMinusOne];
        	pixels[iMinusOne + img->w *bpp] = pixels[iMinusOne];
        	//pixels[iMinusOne + img->w*2] = pixels[iMinusOne];  //vertical copying of pixels
        }
 
 
}
*/

void ImageProc::linInt(SDL_Surface *img) {}

SDL_Surface* ImageProc::minimapScale(SDL_Surface *input) {
    SDL_Surface *output;
    int bytesPerPixel = input->format->BytesPerPixel;
    output = SDL_CreateRGBSurface (SDL_SWSURFACE, (input->w/4),(input->h/ 4), bytesPerPixel *8, 0xff, 0xff, 0xff, 0);
    output->format->Rmask = input->format->Rmask;
    output->format->Gmask = input->format->Gmask;
    output->format->Bmask = input->format->Bmask;
    output->format->Amask = input->format->Amask;
    if( input->format->palette != NULL ) {
        SDL_SetColors(output, input->format->palette->colors, 0, input->format->palette->ncolors);
    }
    Uint8 *pixelsIn = (Uint8 *)input->pixels;
    Uint8 *pixelsOut = (Uint8 *)output->pixels;
    //put the pixels of the input image in the output image with a space of 1
    //pixel between every pixel (horizontal as well as vertical)
    int i= 0;
    int j = 0;
    int lineCounter = 0;
    int iFactor = 4 * bytesPerPixel;
    int limit = ((input->w * input->h) * bytesPerPixel) - iFactor;
    for(i = 0; i <  limit;i+=iFactor, j+=bytesPerPixel, lineCounter++) {
        if(lineCounter == 6 ) {
            lineCounter= 0;
            i+= input->w * bytesPerPixel * 3;
        }
        pixelsOut[j] = pixelsIn [i];
        pixelsOut[j+1] = pixelsIn[i+1];
    }
    return output;
}
