#include <stdlib.h>
#include "minigfx.h"
#include "shpimage.h"
#include "font.h"

#define MAX(x,y) (((x)>=(y))?(x):(y))

MiniGFX::MiniGFX() {
    SDL_WM_SetCaption("SHP Viewer", NULL);
    screen = SDL_SetVideoMode(640, 480, 16, SDL_HWSURFACE|SDL_DOUBLEBUF);
    if (screen == NULL) {
        fprintf(stderr, "Unable to init videomode: %s\n", SDL_GetError());
        exit(1);
    }
}

MiniGFX::~MiniGFX() {
}

void MiniGFX::draw(SHPImage* shp, Font* fnt, Uint32 frame,Uint16 max) {
    SDL_Surface* image, *dummy;
    shp->getImage(frame,&image,&dummy,0);
    delete dummy;
    draw_it(image,fnt,frame,shp->getWidth(),shp->getHeight(),max);
}

void MiniGFX::draw(TemplateImage* tem, Font* fnt, Uint32 frame,Uint16 max) {
    SDL_Surface* image;
    image = tem->getImage(frame);
    draw_it(image,fnt,frame,image->w,image->h,max);
}

void MiniGFX::draw(Dune2Image* d2i, Font* fnt, Uint32 frame,Uint16 max) {
    SDL_Surface* image;
    image = d2i->getImage(frame);
    draw_it(image,fnt,frame,image->w,image->h,max);
}

void MiniGFX::draw_it(SDL_Surface* img, Font *fnt, Uint32 frame, Uint16 width, Uint16 height, Uint16 max) {
    SDL_Rect dest;

    static Uint32 blackpix = SDL_MapRGB(screen->format,0, 0, 0);
    static Uint32 whitepix = SDL_MapRGB(screen->format,0xff, 0xff, 0xff);

    char textmsg[1024];

    dest.x = 0;
    dest.y = 0;

    dest.w = 640;
    dest.h = 480;
    SDL_FillRect(screen, &dest, blackpix);

    dest.x += 25;

    dest.w = width;
    dest.h = height;
    dest.x = MAX(320-dest.w,0);
    dest.y = MAX(240-dest.h,0);

    SDL_BlitSurface(img,NULL,screen,&dest);
    delete img;

    sprintf(textmsg, "Frame %d / %d", frame,max);
    fnt->drawString(textmsg, screen, 0, 0);

    SDL_Flip(screen);
}

